/*
 * CenterAddKMeans.java
 *
 * Created on 15. Januar 2007, 17:34
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package clusterer;
import java.util.Random;
import compr.DataObject;
import visu.VisuInfo;

/**
 * Add centers as long as the costs are improving. Distance-based k-Means and gaussCost only.
 * @author boehm
 */
public class CenterAddKMeans {
    DataObject[] data;
    int numCenters;
    int dim;
    double[][] centers; //centers as row vectors;
    double[] stdv; //standard deviations of the clusters
    int[] numObj; //number of objects per cluster for costID
    double[] distances; //distances of the points to the cluster centers: The point with the maximum distance is the next center to add.
    int precision;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    
    
    /** Creates a new instance of CenterAddKMeans */
    public CenterAddKMeans(DataObject[] data, int precision) {
        this.data = data;
        this.precision =  precision;
        dim = data[0].coord.length;
        numCenters = 2;
        centers = new double[2][dim];
        numObj = new int[2];
        distances = new double[data.length];
        stdv = new double[numCenters];
    }
    
    public void run(){
        randomInit();
        boolean converged = false;
        while(!converged){
            converged = assignPoints();
            update();
        }
        double costOld = cost();
        System.out.println("costOld " + costOld + " " + numCenters + " centers.");
        double costNew = 0.0;
        boolean improvement = true;
        int iter = 0;
        while(improvement){
            //while(iter < 2){
            addCenter();
            converged = false;
            while(!converged){
                converged = assignPoints();
                update();
            }
            boolean emptyCluster = false;
            for(int i = 0; i < numObj.length; i++)
                if(numObj[i] == 0)
                    emptyCluster = true;
            if(emptyCluster){
                System.out.println("EMPTY CLUSTER");
                for(int i = 0; i < numObj.length; i++)
                    if(numObj[i] == 0)
                        replaceEmptyWithRandom(i);
                converged = false;
                while(!converged){
                    converged = assignPoints();
                    update();
                }
            }
            costNew = cost();
            System.out.println("costNew " + costNew + " " + numCenters + " centers.");
            if(costNew < costOld){
                improvement = true;
                costOld = costNew;
            } else
                improvement = false;
            iter++;
        }
        System.out.println("done.");
        VisuInfo vi = new VisuInfo(data, "data");
        vi.setSize(600,600);
        vi.setLocation(200,0);
        vi.setVisible(true);
    }
    
    /**
     * Random initialization with 2 clusters.
     */
    public void randomInit(){
        Random r = new Random(15);
        DataObject firstCenter = data[r.nextInt(data.length)];
        DataObject secondCenter = data[r.nextInt(data.length)];
        for(int i = 0; i < dim; i++)
            centers[0][i] = firstCenter.coord[i];
        for(int i = 0; i < dim; i++)
            centers[1][i] = secondCenter.coord[i];
    }
    
    /**
     * Assign points to closest center according to euclidean distance.
     */
    public boolean assignPoints(){
        boolean clusterChanged = false;
        for(int i = 0; i < data.length; i++){
            double minDist = Double.MAX_VALUE;
            int minIndex = -1;
            for(int j = 0; j < numCenters; j++){
                double[] mean = new double[dim];
                for(int k = 0; k < dim; k++)
                    mean[k] = centers[j][k];
                double aktDist = distance(data[i], mean);
                if(aktDist < minDist){
                    minDist = aktDist;
                    minIndex = j;
                }
            }
            if(data[i].clusterID != minIndex)
                clusterChanged = true;
            data[i].clusterID = minIndex;
            numObj[minIndex]++;
            distances[i] = minDist;
        }
        return clusterChanged;
    }
    
    /**
     * Update mean coordinates and stdvs of the clusters.
     */
    public void update(){
        for(int i = 0; i < numCenters; i++){
            double[] sumCoord = new double[dim]; //for mean
            int counter = 0;
            for(int j = 0; j < data.length; j++){
                if(data[j].clusterID == i){
                    for(int k = 0; k < dim; k++)
                        sumCoord[k] += data[j].coord[k];
                }
                counter++;
            }
            for(int k = 0; k < dim; k++){
                sumCoord[k]/= counter;
                centers[i][k] = sumCoord[k];
            }
        }
        for(int i = 0; i < numCenters; i++){
            double[] sumDev = new double[dim]; //for stdv
            int counter = 0;
            for(int j = 0; j < data.length; j++){
                if(data[j].clusterID == i){
                    for(int k = 0; k < dim; k++)
                        sumDev[k] += (centers[i][k] - data[j].coord[k]) * (centers[i][k] - data[j].coord[k]) ;
                }
                counter++;
            }
            for(int k = 0; k < dim; k++){
                sumDev[k]/= counter;
                stdv[k] = Math.sqrt(sumDev[k]);
            }
        }
    }
    
    private void addCenter(){
        //determine the next center as the most expensive point
        double maxDist = 0.0;
        int maxIndex = -1;
        for (int i = 0; i < distances.length; i++){
            if(distances[i] > maxDist){
                maxDist = distances[i];
                maxIndex = i;
            }
        }
        //add this point to centers.
        numCenters++;
        double[][] newCenters = new double[numCenters][dim];
        for(int i = 0; i < newCenters.length-1; i++)
            for(int j = 0; j < dim; j++)
                newCenters[i][j] = centers[i][j];
        for(int j = 0; j < dim; j++)
            newCenters[numCenters-1][j] = data[maxIndex].coord[j];
        centers = newCenters;
        stdv = new double[numCenters];
        numObj = new int[numCenters];
        distances = new double[data.length];
    }
    
    private void replaceEmptyWithRandom(int centerIndex){
        double[][] newCenters = new double[numCenters][dim];
        for(int i = 0; i < numCenters; i++){
            if(i != centerIndex)
                for(int j = 0; j < dim; j++)
                    newCenters[i][j] = centers[i][j];
        }
        Random r = new Random();
        DataObject co = data[r.nextInt(data.length)];
        for(int j = 0; j < dim; j++)
            newCenters[centerIndex][j] = co.coord[j];
    }
    
    
    /**
     * Determine cost: Coding cost for the points w.r.t. current cluster centers and parameter costs.
     * @return costs for current configuration.
     */
    public double cost(){
        double coordinateCost = 0;
        double costID = 0;
        for(int i = 0; i < data.length; i++)
            coordinateCost += coordinateCost(data[i]);
        System.out.println("---determining costs---");
        System.out.println("coordinateCost  " + coordinateCost);
        for(int i = 0; i < numCenters; i++)
            costID += costID(numObj[i]);
        System.out.println("costID " + costID);
        return coordinateCost+costID;
    }
    
    
    private double costID(int numObj){
        if(numObj == 0)
            return 0;
        else
            return numObj * Math.log((double)data.length/(double)numObj)/Math.log(2);
    }
    
    
    private double coordinateCost(DataObject d){
        double cost = 0.0;
        for(int i = 0; i < dim; i++)
            cost += gaussCost(d.coord[i], centers[d.clusterID][i], stdv[i], precision);
        return cost;
    }
    
    private double gaussCost(double x, double mean, double std, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double var = Math.max(std*std, Double.MIN_VALUE) ;
        return Math.max(0.0,(0.5*Math.log(2*Math.PI*var)+precision*ln10+Math.pow(x-mean,2)/2/var)/ln2 );
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    /**
     * Euclidean Distance between object and mean.
     * @param d
     * @param mean
     * @return
     */
    public double distance(DataObject d, double[] mean){
        double sum = 0.0;
        for (int i = 0; i < d.coord.length; i++){
            double dist = ((d.coord[i] - mean[i])*(d.coord[i] - mean[i]));
            sum = sum + dist;
        }
        return Math.sqrt(sum);
    }
    
}
