/*
 * FastICA.java
 *
 * Created on 10. November 2005, 12:28
 *
 */

package ica;
import Jama.*;
import java.util.*;

/**
 * The implementation of the FastICA algorithm
 * This package needs the Jama classes for Matrix 
 * and EigenvalueDecomposition. The usage is similar
 * to EigenvalueDecomposition. The actual computation
 * is performed in the costructor. Then, some methods
 * to access the result in various ways are applicable.
 * @author boehm
 */
public class FastICA {
    
    private Matrix v ;
    private Matrix vInv ;
    private Matrix transf ;
    private Matrix retransf ;
    private Matrix data ;
    private Matrix whData ;
    private int numObj ;
    private int dim ;
    private int alliter ;
    
    /**
     * Performs the fast ICA algorithm on the data matrix (objects in rows)
     * @param data Matrix that contains the objects in rows and the attributes in columns
     * @param symmetric determines whether to use symmetric (true) or Gram-Schmitt orthonormalization
     * @param whiten determines whether to whiten (true) the data before processing. Whitening means transforming the data to zero means and unit variance. Otherwise, the data is centered only (zero means).
     */
    public FastICA(Matrix data, boolean symmetric, boolean whiten) {
        this.data = data ;
        numObj= data.getRowDimension() ;
        dim = data.getColumnDimension() ;
        Random r = new Random();
        
        if (whiten)
            whData = whitenData(centerData(data)) ;
        else
            whData = centerData(data) ;
        
        double[][] w = new double[dim][dim] ;
        alliter = 0 ;
        double sum = 0.0 ;
        
        if (symmetric) {
            for (int n=0 ; n<dim ; n++)
                for (int i=0; i<dim ; i++)
                    w[n][i] = r.nextDouble() ;
            w = symmetricOrthonorm(w) ;
            do {
                double[][] old_w = w.clone() ;
                for (int n=0 ; n<dim ; n++)
                    w[n] = optimize(w[n]) ;
                w = symmetricOrthonorm(w) ;
                sum = 0.0 ;
                for (int n=0 ; n<dim ; n++)
                    for (int i=0; i<dim ; i++)
                        sum += w[n][i] * old_w[n][i] ;      
                alliter ++ ;
            } while (alliter<100 && sum<dim*0.9999) ;
            
            // System.out.println("Iterations: "+alliter) ;
            
        } else {
            
            for (int n=0 ; n<dim ; n++) {
                
                for (int i=0; i<dim ; i++)
                    w[n][i] = r.nextDouble() ;
                
                normalize(w[n]) ;
                
                double [] old_w ;
                double [] wUnorth ;
                int iter = 0 ;
                
                do {
                    old_w = w[n] ;
                    w[n] = optimize(w[n]) ;
                    // deflationary orthogonalize w[n] w.r.t. vectors w[0]..w[n-1]
                    
                    wUnorth = w[n].clone();
                    
                    for (int m=0 ; m<n ; m++) {
                        double scalar = 0.0 ;
                        for (int j=0 ; j<dim ; j++)
                            scalar += wUnorth[j] * w[m][j] ;
                        for (int j=0 ; j<dim ; j++)
                            w[n][j] -= scalar * w[m][j] ;
                    }
                    normalize(w[n]) ;
                    sum = 0.0 ;
                    for (int i=0; i<dim ; i++)
                        sum += w[n][i] * old_w[i] ;
                    iter ++ ;
                } while (Math.abs(sum)<0.9999 && iter<100);
                
                alliter += iter ;
            }
            
        }
        
        // retransform the matrix w to the original (not whitened) data space
        // w' = w' * retransform
        
        if (whiten) {
            Matrix h = new Matrix(w) ;
            v = h.times(retransf) ;
            vInv = transf.times(h.transpose()) ;
            /* Normalization of the matrix:
             * This just turned out to be unuseful as the non-orthogonal matrix
             * does not preserve any distances anyway. If any, it would rather
             * be useful to normalize vInv
            for (int m=0 ; m<dim ; m++) {
                double sum = 0.0 ;
                Matrix hMat = v.getMatrix(0,dim-1,m,m) ;
                double norm = Math.sqrt(hMat.transpose().times(hMat).get(0,0)) ;
                System.out.println(m+".Norm: "+norm) ;
                for (int i=0; i<dim ; i++)
                    v.setMatrix(0,dim-1,m,m,hMat.times(1.0/norm)) ;
                vInv.setMatrix(m,m, 0,dim-1,vInv.getMatrix(m,m,0,dim-1).times(norm)) ;
            }
             */
        } else {
            v = new Matrix(w) ;
            vInv = v.transpose() ;
        }
        whData = null;
    }
    
    /**
     * The same as above with symmetric and whitening = true
     * @param data Matrix that contains the objects in rows and the attributes 
     * in columns
     */
    public FastICA(Matrix data) {
        this(data, true, true) ;
    }
    
    /**
     * the same as above but data is a double[][] array rather than a matrix
      * @param data array that contains the objects in rows and the attributes in columns
    */
    public FastICA(double[][] data) {
        this(new Matrix(data), true, true) ;
    }
    
    /**
     * Get the matrix containing the main directions
     * @return Contains the vectors of main directions in column vectors
     */
    public Matrix getV() {
        return v ;
    }
    
    /**
     * Returns the inverse matrix of v, used to transform the data
     * @return Inverse matrix of getV()
     */
    public Matrix getVI() {
        return vInv ;
    }
    
    /**
     * Transforms any data matrix according to the main directions which have 
     * been found using the constructor. The data matrix can be the same as used 
     * in the constructor but does not have to. Note that ICA is *not* done new 
     * if there is another data matrix.
     * @param dat contains objects in rows and attributes in columns
     * @return the transformed data matrix dat*getVI()
     */
    public Matrix transform(Matrix dat) {
        return dat.times(vInv) ;
    }
    
    /**
     * The same as above, but with the original data matrix. Note that the data 
     * matrix is not cloned by the constructor, so if the data matrix has been 
     * changed meanwhile, the changed version is used
     * @return data*getVI()
     */
    public Matrix transform() {
        return data.times(vInv) ;
    }
    
    /**
     * Number of iterations which were required to do fast ICA computation
     * @return overall number of iterations
     */
    public int getNumIter() {
        return alliter ;
    }
    
    private Matrix centerData(Matrix data) {
        Matrix result = new Matrix(numObj, dim) ;
        
        for (int i=0 ; i<dim ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sum += data.get(j,i) ;
            sum /= numObj ;
            for (int j=0 ; j<numObj ; j++)
                result.set(j,i,data.get(j,i) - sum) ;
        }
        
        return result ;
    }
    
    private Matrix covar(Matrix data) {
        return data.transpose().times(data).times(1.0/data.getRowDimension()) ;
    }
    
    private Matrix whitenData(Matrix data) {
        EigenvalueDecomposition e = new EigenvalueDecomposition(covar(data));
        Matrix evm = e.getV();
        Matrix ewm = e.getD();
        transf = new Matrix(dim, dim) ;
        retransf = new Matrix(dim, dim) ;
        for (int i=0; i<dim ; i++) {
            double diag = Math.sqrt(Math.min(Math.max(ewm.get(i,i), 1.0e-6), 1.0e+6) ) ;
            for (int j=0 ; j<dim ; j++) {
                double h = evm.get(j,i) ;
                transf.set(j, i, h / diag) ;
                retransf.set(i, j, h * diag) ;
            }
        }
        return data.times(transf) ;
    }
    
    private double [] optimize(double [] w) {
        double [] result = w.clone() ;
        double eLeft[] = new double[dim] ;
        double eRight = 0.0 ;
        for (int i=0 ; i<numObj ; i++) {
            double scalar = 0.0 ;
            for (int j=0 ; j<dim ; j++)
                scalar += w[j] * whData.get(i,j) ;
            double g = g_func(scalar) / numObj ;
            for (int j=0 ; j<dim ; j++)
                eLeft[j] += g * whData.get(i,j) ;
            eRight += g_slope(scalar) / numObj ;
            
        }
        for (int j=0 ; j<dim ; j++)
            result[j] = eLeft[j] - eRight * result[j] ;
        return result ;
    }
    
    private void normalize(double [] w) {
        double sum = 0.0 ;
        for (int i=0; i<dim ; i++)
            sum += w[i]*w[i] ;
        sum = Math.sqrt(sum) ;
        for (int i=0; i<dim ; i++)
            w[i] /= sum ;
    }
    
    private double[][] symmetricOrthonorm(double[][] w) {
        Matrix wm = new Matrix(w) ;
        EigenvalueDecomposition e = new EigenvalueDecomposition(wm.times(wm.transpose()));
        Matrix d = e.getD() ;
        for (int i=0 ; i<dim ; i++)
            d.set(i,i, 1.0/Math.sqrt(d.get(i,i))) ;
        double [][] result = e.getV().times(d).times(e.getV().transpose()).times(wm).getArrayCopy() ;
        return result ;
    }
    
    /**
     * Function used in the fastICA algorithm to estimate negentropy. Can be 
     * replaced by a different function (maybe by overriding)
     * @return result of the function at position x
     * @param x The position x of function evaluation
     */
    public double g_func(double x) {
        return Math.tanh(x) ;
    }
    
    /**
     * Derivative of the above function
     * @param x The position x of function evaluation
     * @return The derivative of the function at position x
     */
    public double g_slope(double x) {
        double h=Math.tanh(x) ;
        return 1.0-h*h ;
    }
}
