/*
 * TopDownKM.java
 *
 * Created on 18. April 2006, 15:19
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */
package oci;

import java.io.IOException;
import java.util.*;
import Jama.*;
import com.jmatio.io.MatFileWriter;
import com.jmatio.types.MLDouble;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Class to support both IcaSplit and bisecting K-Means (PCASplit). 
 *MOST IMPORTANT THINGS TO KNOW ABOUT THE IMPLEMENTATION:
 *- Method run implements the algorithm with steps: splitting, outlier filtering and merging. For details see annotations
 *- handling of data: current clusters are stored in Vectors clusters and noiseClusters, currend cluster IDs in array aktClusterID
 *  best solution is stored in bestClusters, bestNoiseClusters.
 *- coding costs: According to MDL the total costs of a clustering are: ID costs (for each object) + parameter costs (for each cluster) + coding costs (for each object) 
 *- ID costs: not stored. Are computed in methods currentCost() and savedCost() when comparing clusterings with different numbers of clusters
 *- parameterCosts: fixed cost for one parameter: 32 bits. Parameter costs are determined in Class IcaCost, where the best model for the clusters is determined.
 *- coding costs: are also computed in IcaCost. The field .cost of IcaCluster contains parameter costs for the cluster and coding costs for its associated objects 
 *- During the filtering step implemented by the method finalFiltering(), no parameter costs for the noise clusters are added (to make noise filtering possible)
 *- Parameter costs for noise are added before merging in the method greedyMerge()
 *
 * @author claudia
 */
public class OCI {

    DataObject[] data;
    Vector<IcaCluster> clusters; //current clusters
    Vector<IcaCluster> noiseClusters;
    IcaCluster[] bestClusters; //cheapest clusters
    IcaCluster[] bestNoiseClusters;
    int[] aktClusterID; //current cluster IDs;
    IcaCluster[] bestClustersICA; //current best clusters obtained with Ica
    IcaCluster[] bestNoiseClustersICA;
    IcaCluster[] bestClustersPCA; //current best clusters obtained with KM
    IcaCluster[] bestNoiseClustersPCA;
    int[] idsICA; //current best clusterIDs bisecting K-Means
    int[] idsPCA; //current best clusterIDs IcaSplit
    DB db;
    double bestCost;
    int iter;
    int numObj;
    int dim;
    int maxClusterID;      //Max ID of the clusters (not noise!)
    int numTry = 0;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    static int minNumObj = 20; //minimum number of objects in a cluster for ica

    /** Creates a new instance of OCI */
    public OCI(DataObject[] data) {
        this.data = data;
        db = new DB(data);
        numObj = data.length;
        iter = 0;
        dim = data[0].coord.length;
        maxClusterID = 0;
        bestCost = 0.0;
        clusters = new Vector<IcaCluster>();
        noiseClusters = new Vector<IcaCluster>();
        idsPCA = new int[numObj];
        idsICA = new int[numObj];
    }

    /**
     * Filter clusters after splitting with bisecting K-Means.
     * COSTS: no parameter costs added here, this is done for the noise objects in the merge method.
     */
    private void finalFiltering() {
        System.out.println("--- finalFiltering ---");
        Vector<IcaCluster> filteredClusters = new Vector<IcaCluster>();
        Vector<IcaCluster> filteredNoise = new Vector<IcaCluster>();
        for (int i = 0; i < bestClusters.length; i++) {
            System.out.println("filtering cluster " + bestClusters[i].ID + " numObj " + bestClusters[i].numObj);
            DataObject[] o = getObjects(bestClusters[i]);
            System.out.println("o.length " + o.length);
            IcaCluster[] result = getBestFilteredClusters(o, true, true);
            result[0].ID = bestClusters[i].ID;
            result[1].ID = bestClusters[i].ID + 1;
            //adjust clusterIDs
            for (int j = 0; j < aktClusterID.length; j++) {
                if (aktClusterID[j] == 1) {
                    data[o[j].number].clusterID = result[1].ID;
                }
                if (aktClusterID[j] == 0) {
                    data[o[j].number].clusterID = result[0].ID;
                }
            }
            filteredClusters.add(result[0]);
            filteredNoise.add(result[1]);
        }
        IcaCluster[] filteredC = new IcaCluster[filteredClusters.size()];
        for (int i = 0; i < filteredC.length; i++) {
            filteredC[i] = (IcaCluster) filteredClusters.elementAt(i);
        }
        bestClusters = filteredC;
        IcaCluster[] filteredN = new IcaCluster[filteredNoise.size()];
        for (int i = 0; i < filteredN.length; i++) {
            filteredN[i] = (IcaCluster) filteredNoise.elementAt(i);
        }
        bestNoiseClusters = filteredN;
    }

    private void testMethodForCodingCostObjects() {
        for (int i = 0; i < bestClusters.length; i++) {
            double sumClusterWiseCodingCosts = 0.0;
            for (int j = 0; j < bestClusters[i].codingCostObjects.length; j++) {
                sumClusterWiseCodingCosts += bestClusters[i].codingCostObjects[j];
            }
            if (sumClusterWiseCodingCosts - bestClusters[i].codingCost > 0.00001) {
                System.err.println("Wrong cluster wise coding cost: cluster " + bestClusters[i].ID);
            } else {
                System.out.println("---CodingCostsObjects: ok-----");
            }
        }

    }

    private void testMethodForAssignedObjects() {
        System.out.println("-----testMethodForAssignedObjects-----");
        int[] assigned = new int[data.length];
        for (int i = 0; i < bestClusters.length; i++) {
            for (int j = 0; j < bestClusters[i].assignedObjects.length; j++) {
                assigned[bestClusters[i].assignedObjects[j]]++;
            }
        }
        for (int i = 0; i < assigned.length; i++) //must be all 1
        {
            if (assigned[i] != 1) {
                System.err.println("error in object assigment.");
            }
        }

    }

    /**
     * MODIFIED: returns filtered or unfiltered clusters. To enable no filtering during splitting.
     * COSTS: Parameter costs for the cluster are added during splitting. For final filtering parameter costs for the cluster are added. Parameter costs for noise clusters are added in merge method.
     * @param o
     * @param filter if noise should be filtered
     * @param verbose
     * @return Cluster[2]: result[0]: filtered cluster, result[1]: filtered noise, can be empty (numObj = 0).
     */
    private IcaCluster[] getBestFilteredClusters(DataObject[] o, boolean filter, boolean verbose) {
        if (verbose) {
            System.out.println("---getBestFilteredClusters, filter: " + filter);
        }
        if (filter) {
            if (o.length < minNumObj) {
                if (verbose) {
                    System.out.println("Cluster too small for filtering");
                }
                double paramCost = dim * 32 * 2;
                IcaCluster[] result = new IcaCluster[2];
                IcaCost ic = new IcaCost(dim);
                result[0] = ic.mergedClusterNoRot(o);
                result[0].parameterCost = paramCost;
                result[0].cost += paramCost;
                result[1] = new IcaCluster(1);
                //clusterID
                int[] ids = new int[o.length];
                for (int i = 0; i < ids.length; i++) {
                    ids[i] = 0;
                }
                aktClusterID = ids;
                return result;
            }
            IcaCost ic = new IcaCost(o);
            IcaCluster[] result = ic.selectBestFilteredCluster();
//            //add parameter costs
//            double paramCostCluster = paramCost; //add parameter costs for cluster
//            double paramCostNoise = 0.0;
//            result[0].cost += paramCostCluster;
//            result[1].cost += paramCostNoise;
//            if(verbose){
//                System.out.println("coding cost cluster: " + result[0].cost + " including " + paramCostCluster + " parameter costs.");
//                System.out.println("coding cost noise: " + result[1].cost + " including " + paramCostNoise + " parameter costs.");
//            }
            int[] ids = ic.getClusterID();
            aktClusterID = ids;
            return result;
        } else { //all objects in cluster
            if (o.length < minNumObj) {
                if (verbose) {
                    System.out.println("Cluster too small for ica");
                }
                double paramCost = dim * 32 * 2;
                IcaCluster[] result = new IcaCluster[2];
                IcaCost ic = new IcaCost(dim);
                result[0] = ic.mergedClusterNoRot(o);
                result[0].parameterCost = paramCost;
                result[0].cost += paramCost;
                result[1] = new IcaCluster(1);
                //clusterID
                int[] ids = new int[o.length];
                for (int i = 0; i < ids.length; i++) {
                    ids[i] = 0;
                }
                aktClusterID = ids;
                aktClusterID = ids;
                return result;
            }

            int[] ids = new int[o.length];
            for (int i = 0; i < ids.length; i++) {
                ids[i] = 0;
            }
            aktClusterID = ids;
            IcaCost ic = new IcaCost(dim);
            IcaCluster[] result = new IcaCluster[2];
            // 20 is the minimal kNN size, i.e. for all NN larger than 20 the cost function is calculated an saved
            KNN knn = new KNN(20);
            knn.getAllNN(o, dim);
            result[0] = ic.selectBestCluster(o);
            result[0].ID = 0;
            //result[0].cost += paramCost;
            result[0].numObj = o.length;
            result[1] = new IcaCluster(1);
            result[1].numObj = 0;
            return result;
        }
    }

    public void outlier() {
        outlierSplit();
        //finalFiltering();
        double[][] cost = greedyMergeForOutlier(true);
        assignPoints(cost);
        VisuOutlier vi = new VisuOutlier(data,  "result", 50);
        vi.setSize(600, 600);
        vi.setLocation(200, 0);
        vi.setVisible(true);
        
       
    }

    public void outlierSplit() {
        //split down as in OCI
        double splitMinCost = Double.MAX_VALUE; //cost minimum during splitting
        IcaCluster[] splitMinClusters = new IcaCluster[1]; //clusters with minimum cost achieved only by splitting without filtering
        IcaCluster[] result = getBestFilteredClusters(data, false, false);
        int[] splitIds = new int[data.length];  //save clusterIDs after split
        for (int i = 0; i < data.length; i++) {
            data[i].clusterID = aktClusterID[i];
        }
        bestClusters = new IcaCluster[1];
        bestClusters[0] = result[0];
        bestNoiseClusters = new IcaCluster[1];
        bestNoiseClusters[0] = result[1];
        double costOneCluster = currentCost();
        splitMinCost = costOneCluster;
        double minCost = costOneCluster; //minimum cost
        boolean improvement = true;
        int aktTry = 0;
        IcaCluster[] lastLocalMinClusters = new IcaCluster[1];        //save Clusters and ClusterIDs of the last local minimum
        IcaCluster[] lastLocalMinNoise = new IcaCluster[1];
        int[] lastLocalMinIDs = new int[numObj];
        IcaCluster[] minClusters = new IcaCluster[1];        //save Clusters and ClusterIDs of the global cost minimum
        IcaCluster[] minNoise = new IcaCluster[1];
        int[] minIDs = new int[numObj];
        minClusters = bestClusters;
        minNoise = bestNoiseClusters;
        for (int i = 0; i < data.length; i++) {
            minIDs[i] = data[i].clusterID;
        }
        while (improvement || aktTry < numTry) {
            //System.out.println("-------------------------------------");
            double costUnsplit = currentCost();
            int numClUnsplit = numClustersNumNoiseClusters()[0];
            int numNUnsplit = numClustersNumNoiseClusters()[1];
            //System.out.println("SYMMETRIC SPLIT NUMBER: " + iter + " CURRENT COST " + costUnsplit);
            getCostInfoShort();
            improvement = symmetricSplit(aktTry);
            if (improvement) {
                minCost = currentCost();
                //new cost minimum found
                if (minCost < splitMinCost) {
                    splitMinCost = minCost;
                    splitMinClusters = bestClusters;
                    for (int i = 0; i < splitIds.length; i++) {
                        splitIds[i] = data[i].clusterID;
                    }
                    minClusters = bestClusters;
                    minNoise = bestNoiseClusters;
                    for (int i = 0; i < numObj; i++) {
                        minIDs[i] = data[i].clusterID;
                    }
                    //System.out.println("GLOBAL IMPROVEMENT: " + " CURRENT COST " + minCost + " COST OLD: " + costUnsplit);
                    getCostInfoShort();
                } else //in this case: improvement for at least one cluster but no overall improvement
                {
                    //System.out.println("IMPROVEMENT FOR AT LEAST ONE CLUSTER: " + minCost + " COST OLD: " + costUnsplit);
                    getCostInfoShort();
                }
                lastLocalMinClusters = bestClusters;
                lastLocalMinNoise = bestNoiseClusters;
                for (int i = 0; i < numObj; i++) {
                    lastLocalMinIDs[i] = data[i].clusterID;
                }
                aktTry = 0;
            } else { //in this case: improvement for no cluster
                aktTry++;
                //System.out.println("NO GLOBAL IMPROVEMENT: CURRENT COST " + currentCost()/*currentCostWithoutIDCosts()*/ + " COST OLD: " + costUnsplit + " TRYING FOR ANOTHER " + (numTry - aktTry) + " TIMES");
                IO ea = new IO();
//                ea.printResult(data);
//                System.out.println("m");
            }
            //System.out.println("OLD: " + numClUnsplit + " clusters, " + numNUnsplit + " noiseClusters.");
            //System.out.println("NEW: " + numClustersNumNoiseClusters()[0] + " clusters " + numClustersNumNoiseClusters()[1] + " noiseClusters.");
            //System.out.println("-------------------------------------");
            getCostInfoShort();
            iter++;
        }
        //System.out.println("done. " + iter + " iterations");

        //adjust clusters and IDs as from the last local minimum
        if (minClusters.length > 1 || minNoise.length > 1) {
            bestClusters = minClusters;
            bestNoiseClusters = minNoise;
            for (int i = 0; i < data.length; i++) {
                data[i].clusterID = minIDs[i];
            }
        //System.out.println("BEST: " + bestClusters.length + " clusters ");
        }
        //continue splitting each cluster until minclustersize is reached. split in each step one cluster with the least cost detoriation.
        boolean splitPerformed = true;
        int numSplit = 0;
        while (splitPerformed) {
            splitPerformed = splitForOutlier();
            //testMethodForCodingCostObjects();
            getCostInfoShort();
        }
        //testMethodForAssignedObjects();
        int[] currentIDs = new int[data.length];
        for (int i = 0; i < currentIDs.length; i++) {
            currentIDs[i] = data[i].clusterID;
        }




        Clustering c = new Clustering(-1, bestClusters, currentIDs);
        //write Serializable Objects

        IO ea = new IO();
        ea.writeClusteringAsSerializable(c, "mist");
        System.out.println("m");

    //ea.printClusters(data, 4);
    //ea.printClusters(data, 2);
    //ea.printClusters(data, 0);
    //ea.printClusters(data, 10);

    //print clustering result
    //ea.printResult(data);
    //ea.setClusterID(data); //change clusterID for display



    }

    /**
     * implements OCI
     **/
    public void run() {
        double splitMinCost = Double.MAX_VALUE; //cost minimum during splitting
        IcaCluster[] splitMinClusters = new IcaCluster[1]; //clusters with minimum cost achieved only by splitting without filtering
        IcaCluster[] result = getBestFilteredClusters(data, false, false);
        int[] splitIds = new int[data.length];  //save clusterIDs after split
        for (int i = 0; i < data.length; i++) {
            data[i].clusterID = aktClusterID[i];
        }
        bestClusters = new IcaCluster[1];
        bestClusters[0] = result[0];
        bestNoiseClusters = new IcaCluster[1];
        bestNoiseClusters[0] = result[1];
        double costOneCluster = currentCost();
        splitMinCost = costOneCluster;
        double minCost = costOneCluster; //minimum cost
        boolean improvement = true;
        int aktTry = 0;
        IcaCluster[] lastLocalMinClusters = new IcaCluster[1];        //save Clusters and ClusterIDs of the last local minimum
        IcaCluster[] lastLocalMinNoise = new IcaCluster[1];
        int[] lastLocalMinIDs = new int[numObj];
        IcaCluster[] minClusters = new IcaCluster[1];        //save Clusters and ClusterIDs of the global cost minimum
        IcaCluster[] minNoise = new IcaCluster[1];
        int[] minIDs = new int[numObj];
        minClusters = bestClusters;
        minNoise = bestNoiseClusters;
        for (int i = 0; i < data.length; i++) {
            minIDs[i] = data[i].clusterID;
        }
        while (improvement || aktTry < numTry) {
            System.out.println("-------------------------------------");
            double costUnsplit = currentCost();
            int numClUnsplit = numClustersNumNoiseClusters()[0];
            int numNUnsplit = numClustersNumNoiseClusters()[1];
            System.out.println("SYMMETRIC SPLIT NUMBER: " + iter + " CURRENT COST " + costUnsplit);
            improvement = symmetricSplit(aktTry);
            if (improvement) {
                minCost = currentCost();
                //new cost minimum found
                if (minCost < splitMinCost) {
                    splitMinCost = minCost;
                    splitMinClusters = bestClusters;
                    for (int i = 0; i < splitIds.length; i++) {
                        splitIds[i] = data[i].clusterID;
                    }
                    minClusters = bestClusters;
                    minNoise = bestNoiseClusters;
                    for (int i = 0; i < numObj; i++) {
                        minIDs[i] = data[i].clusterID;
                    }
                    System.out.println("GLOBAL IMPROVEMENT: " + " CURRENT COST " + minCost + " COST OLD: " + costUnsplit);
                } else //in this case: improvement for at least one cluster but no overall improvement
                {
                    System.out.println("IMPROVEMENT FOR AT LEAST ONE CLUSTER: " + minCost + " COST OLD: " + costUnsplit);
                }
                lastLocalMinClusters = bestClusters;
                lastLocalMinNoise = bestNoiseClusters;
                for (int i = 0; i < numObj; i++) {
                    lastLocalMinIDs[i] = data[i].clusterID;
                }
                aktTry = 0;
            } else { //in this case: improvement for no cluster
                aktTry++;
                System.out.println("NO GLOBAL IMPROVEMENT: CURRENT COST " + currentCost()/*currentCostWithoutIDCosts()*/ + " COST OLD: " + costUnsplit + " TRYING FOR ANOTHER " + (numTry - aktTry) + " TIMES");
                IO ea = new IO();
//                ea.printResult(data);
//                System.out.println("m");
            }
            System.out.println("OLD: " + numClUnsplit + " clusters, " + numNUnsplit + " noiseClusters.");
            System.out.println("NEW: " + numClustersNumNoiseClusters()[0] + " clusters " + numClustersNumNoiseClusters()[1] + " noiseClusters.");
            System.out.println("-------------------------------------");
            iter++;
        }
        System.out.println("done. " + iter + " iterations");

        //adjust clusters and IDs as from the last local minimum
        if (minClusters.length > 1 || minNoise.length > 1) {
            bestClusters = minClusters;
            bestNoiseClusters = minNoise;
            for (int i = 0; i < data.length; i++) {
                data[i].clusterID = minIDs[i];
            }
            System.out.println("BEST: " + bestClusters.length + " clusters ");
        }
        //checkClusterIDs();
        finalFiltering();
        resetClusterIDs();
        //boolean impr = finalSplit();
        //mergeClusters();
        //simpleGreedyMerge(0);
        greedyMerge(0, true);
        minCost = currentCost();
        System.out.println("BEST COST: " + minCost);
        if (splitMinCost < minCost) {
            bestClusters = splitMinClusters;
            for (int i = 0; i < data.length; i++) {
                data[i].clusterID = splitIds[i];
            }
            System.out.println("NO COST IMPROVEMENT DUE TO FILTERING AND MERGING.");
        //reset clusterIDs
        }

        //printInfo();
        GenGauss g = new GenGauss();
        //g.selectivityEstimation(data, bestClusters, 10);
        //printPercentage();
        //evaluatePDF(1110); //cluster 0
        //evaluatePDF(1003); //gut im Vergleich zu EM auf String dir = "3cluster2dnoiseErzeugt.txt";
        //print data of specified clusters
        IO ea = new IO();
        //ea.printClusters(data, 4);
        //ea.printClusters(data, 2);
        //ea.printClusters(data, 0);
        //ea.printClusters(data, 10);

        //print clustering result
        //ea.printResult(data);
        ea.setClusterID(data); //change clusterID for display
        VisuInfo vi = new VisuInfo(data, "result");
        vi.setSize(600, 600);
        vi.setLocation(200, 0);
        vi.setVisible(true);
    }

    private int[] numClustersNumNoiseClusters() {
        int numCl = 0;
        int numN = 0;
        for (int i = 0; i < bestClusters.length; i++) {
            if (bestClusters[i].numObj > 0) {
                numCl++;
            }
        }
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            if (bestNoiseClusters[i].numObj > 0) {
                numN++;
            }
        }
        int[] result = new int[2];
        result[0] = numCl;
        result[1] = numN;
        return result;
    }

    /**
     * Evaluate how well the clustering result fits to the discovered pdfs. Determine the likelihood of the cluster points and how many points actually are in this location.
     */
    public void evaluatePDF(int rp) {
        double windowSize = 2.5;
        int steps = 1000000;
        double[] lb = new double[dim];
        double ub[] = new double[dim];
        double[] overallLb = new double[dim]; //for KL-divergence
        double[] overallUb = new double[dim];
        for (int i = 0; i < dim; i++) {
            overallLb[i] = Double.MAX_VALUE;
            overallUb[i] = -Double.MAX_VALUE;
        }
        GenGauss gg = new GenGauss();
        for (int i = 0; i < dim; i++) {
            lb[i] = data[rp].coord[i] - windowSize;
            ub[i] = data[rp].coord[i] + windowSize;
        }
        //count Objects in window
        int counter = 0;
        for (int i = 0; i < data.length; i++) {
            boolean contained = true;
            for (int j = 0; j < dim; j++) {
                if (data[i].coord[j] > ub[j] || data[i].coord[j] < lb[j]) {
                    contained = false;
                }
                if (data[i].coord[j] < overallLb[j]) {
                    overallLb[j] = data[i].coord[j];
                }
                if (data[i].coord[j] > overallUb[j]) {
                    overallUb[j] = data[i].coord[j];
                }
            }
            if (contained) {
                counter++;
                data[i].clusterID = 100;
            }
        }
        overallLb[0] = 16.0;
        overallLb[1] = 6.0;
        overallUb[0] = 66.0;
        overallUb[1] = 47.0;

        double prob = 0.0;
        for (int i = 0; i < bestClusters.length; i++) {
            System.out.println(bestClusters[i].ID + " " + gg.integral(lb, ub, bestClusters[i].realCenter, new Matrix(bestClusters[i].realRotation).getArrayCopy(), bestClusters[i].beta, steps));
            prob += (double) bestClusters[i].numObj / (double) numObj * gg.integral(lb, ub, bestClusters[i].realCenter, new Matrix(bestClusters[i].realRotation).getArrayCopy(), bestClusters[i].beta, steps);
        }
        System.out.println("prob: " + prob + " " + counter + " objects found. Percentage:  " + (double) counter / (double) numObj);
        System.out.println("EM: " + gg.gaussEvaluation(lb, ub, steps));
        System.out.println("KL-divergence: " + gg.klEvaluation(bestClusters, overallLb, overallUb, steps, numObj));
    }

    //try to split all cluster with size > minClustersize. Try for each cluster icaSplit and KMSplit. In each step accept the split causing the least increasing cost
    private boolean splitForOutlier() {
        boolean splitPerformed = false;
        IcaCluster[] currentClusters = bestClusters.clone();
        double[] improvement_pca = new double[currentClusters.length]; //save improvement of costs with bisecting K-Means
        double[] improvement_ica = new double[currentClusters.length];
        //System.out.println("FORCED SPLIT FOR OUTLIER DETECTION");
//Variables required to determine the best split among all current clusters with ica/km
        double bestImprovement = Double.NEGATIVE_INFINITY; //for the best split
        int[] bestIDs = new int[data.length]; //for the clusterIDs of the best split
        int splitClusterID = -1; //the ID of the currently split cluster
        boolean ica = false; //split with ica or km?
        IcaCluster[] bestClustersOutlierSplit = new IcaCluster[1]; //for the best Clusters
        IcaCluster[] bestNoiseClustersOutlierSplit = new IcaCluster[1];
//END of variable declaration.
        //TEST reset cluster ID after each attempt to split
        int[] aktID = new int[data.length]; //save IDs if split is not successful
        for (int i = 0; i < data.length; i++) {
            aktID[i] = data[i].clusterID;
        }

        for (int i = 0; i < currentClusters.length; i++) {
            for (int j = 0; j < data.length; j++) {
                data[j].clusterID = aktID[j];
                idsPCA[j] = data[j].clusterID;
                idsICA[j] = data[j].clusterID;
            }
            //checkClusterIDs();
            if (currentClusters[i].numObj > minNumObj) {
                try {
                    //TEST
                    //TEST checkClusterIDs(); ok
                    improvement_pca[i] = splitClusterPCAOutlier(i);
                    //TEST checkClusterIDs(); ok
                    maxClusterID = maxClusterID - 2;
                    improvement_ica[i] = splitClusterICAOutlier(i);
                    maxClusterID = maxClusterID - 2;
                    //set bestClusters and clusterIDs to the solution with minimum cost gain

                    if (Math.max(improvement_pca[i], improvement_ica[i]) > bestImprovement) {
                        if (improvement_pca[i] > improvement_ica[i]) {
                            bestImprovement = improvement_pca[i];
                            //bestIDs = idsPCA;
                            for (int j = 0; j < bestIDs.length; j++) {
                                bestIDs[j] = idsPCA[j];
                            }
                            splitClusterID = i;
                            ica = false;
                            bestClustersOutlierSplit = bestClustersPCA;
                            bestNoiseClustersOutlierSplit = bestNoiseClustersPCA;
                        } else {
                            bestImprovement = improvement_ica[i];
                            //bestIDs = idsICA;
                            for (int j = 0; j < bestIDs.length; j++) {
                                bestIDs[j] = idsICA[j];
                            }
                            splitClusterID = i;
                            ica = true;
                            bestClustersOutlierSplit = bestClustersICA;
                            bestNoiseClustersOutlierSplit = bestNoiseClustersICA;
                        }
                    }
                } catch (java.lang.NullPointerException e) {
                    for (int j = 0; j < data.length; j++) {
                        data[j].clusterID = aktID[j];
                    }
                } catch (java.lang.ArrayIndexOutOfBoundsException e) {
                    for (int j = 0; j < data.length; j++) {
                        data[j].clusterID = aktID[j];
                    }
                }
            } else {
                improvement_pca[i] = Double.NEGATIVE_INFINITY;
                improvement_ica[i] = Double.NEGATIVE_INFINITY;
            }
        //TEST
        //System.out.println("check best IDs: ");
        //checkClusterIDs(bestIDs);
        //TEST
        } //i
        //find and perform the best split
        if (bestImprovement > Double.NEGATIVE_INFINITY) {
            maxClusterID = maxClusterID + 2;
            //System.out.println("Split cluster: " + splitClusterID + " cost change: " + bestImprovement);
            bestCost = bestCost + bestImprovement;
            bestClusters = bestClustersOutlierSplit;
            bestNoiseClusters = bestNoiseClustersOutlierSplit;
            for (int i = 0; i < data.length; i++) {
                data[i].clusterID = bestIDs[i];
            }
            //checkClusterIDs();
            splitPerformed = true;
        } else {
        //System.out.println("Splitting finished.");
        }
        return splitPerformed;

    }

    /**
     * MODIFIED. Returns true if either PCASplit or IcaSplit is successful. Sets bestClusters and clusterIDs to the current best solution.
     * @param aktTry number of tried splits.
     * @return boolean: improvement of cost.
     */
    private boolean symmetricSplit(int aktTry) {
        //aktTry = numTry;
        IcaCluster[] currentClusters = bestClusters.clone();
        double[] improvement_pca = new double[currentClusters.length]; //save improvement of costs with bisecting K-Means
        double[] improvement_ica = new double[currentClusters.length];
        boolean[] improvement = new boolean[currentClusters.length];

        for (int i = 0; i < currentClusters.length; i++) {
            int[] aktID = new int[data.length]; //save IDs if split is not successful
            for (int j = 0; j < data.length; j++) {
                aktID[j] = data[j].clusterID;
            }
            //reset idsPCA and idsICA
            for (int j = 0; j < data.length; j++) {
                idsPCA[j] = data[j].clusterID;
                idsICA[j] = data[j].clusterID;
            }
            try {
                improvement_pca[i] = splitClusterPCA(i);
                maxClusterID = maxClusterID - 2;
                improvement_ica[i] = splitClusterICA(i);
                maxClusterID = maxClusterID - 2;
                if (improvement_pca[i] > 0.0 || improvement_ica[i] > 0.0) { //improvement due to PCA or ICASplit
                    //set clusterIDs and bestClusters to the best solution, i.e. to the maximum improvement.
                    improvement[i] = true;
                    maxClusterID = maxClusterID + 2;
                    if (improvement_pca[i] > improvement_ica[i]) {
                        //System.out.println("KM SPLIT");
                        bestCost = bestCost - improvement_pca[i];
                        bestClusters = bestClustersPCA;
                        bestNoiseClusters = bestNoiseClustersPCA;
                        for (int j = 0; j < data.length; j++) {
                            data[j].clusterID = idsPCA[j];
                        }
                    //TEST
                    //checkClusterIDs();
                    //TEST
                    } else {
                        //System.out.println("ICA SPLIT");
                        bestCost = bestCost - improvement_ica[i];
                        bestClusters = bestClustersICA;
                        bestNoiseClusters = bestNoiseClustersICA;
                        for (int j = 0; j < data.length; j++) {
                            data[j].clusterID = idsICA[j];
                        }
                    //TEST
                    //checkClusterIDs();
                    //TEST
                    }
                } else {
                    //System.out.println("SPLIT UNSUCCESSFUL.");
                    for (int j = 0; j < data.length; j++) {
                        data[j].clusterID = aktID[j];
                    }
                //TEST
                //checkClusterIDs();
                //TEST
                }//else
                if (improvement_pca[i] == Double.NEGATIVE_INFINITY && improvement_ica[i] == Double.NEGATIVE_INFINITY) { //cluster too small for splitting or with both PcaSplit and IcaSplit one of the resulting cluster becomes empty.
                    improvement[i] = false;
                    for (int j = 0; j < data.length; j++) {
                        data[j].clusterID = aktID[j];
                    }
                }

            } catch (java.lang.NullPointerException e) {
                improvement[i] = false;
                for (int j = 0; j < data.length; j++) {
                    data[j].clusterID = aktID[j];
                }
            } catch (java.lang.ArrayIndexOutOfBoundsException e) {
                System.err.println("mist");
                improvement[i] = false;
                for (int j = 0; j < data.length; j++) {
                    data[j].clusterID = aktID[j];
                }
            }
        }
        boolean impr = false;
        for (int i = 0; i < improvement.length; i++) {
            if (improvement[i]) {
                impr = true;
            }
        }
        if (!impr) {
            if (aktTry < numTry) {
                //System.out.println("FORCED SPLIT");
                for (int i = 0; i < currentClusters.length; i++) {
                    int[] aktID = new int[data.length]; //save IDs if split is not successful
                    improvement[i] = true;
                    try {
                        improvement_pca[i] = splitClusterPCA(i);
                        maxClusterID = maxClusterID - 2;
                        improvement_ica[i] = splitClusterICA(i);
                        maxClusterID = maxClusterID - 2;
                        //set bestClusters and clusterIDs to the solution with minimum cost gain
                        maxClusterID = maxClusterID + 2;
                        if (improvement_pca[i] > improvement_ica[i]) {
                            // System.out.println("KM SPLIT");
                            bestCost = bestCost - improvement_pca[i];
                            bestClusters = bestClustersPCA;
                            bestNoiseClusters = bestNoiseClustersPCA;
                            for (int j = 0; j < data.length; j++) {
                                data[j].clusterID = idsPCA[j];
                            }
                        } else {
                            //System.out.println("ICA SPLIT");
                            bestCost = bestCost - improvement_ica[i];
                            bestClusters = bestClustersICA;
                            bestNoiseClusters = bestNoiseClustersICA;
                            for (int j = 0; j < data.length; j++) {
                                data[j].clusterID = idsICA[j];
                            }
                        }
                        if (improvement_pca[i] == Double.NEGATIVE_INFINITY && improvement_ica[i] == Double.NEGATIVE_INFINITY) { //cluster too small for splitting or with both PcaSplit and IcaSplit one of the resulting cluster becomes empty.
                            improvement[i] = false;
                            for (int j = 0; j < data.length; j++) {
                                data[j].clusterID = aktID[j];
                            }
                        }
                    } catch (java.lang.NullPointerException e) {
                        improvement[i] = false;
                        for (int j = 0; j < data.length; j++) {
                            data[j].clusterID = aktID[j];
                        }
                    } catch (java.lang.ArrayIndexOutOfBoundsException e) {
                        improvement[i] = false;
                        for (int j = 0; j < data.length; j++) {
                            data[j].clusterID = aktID[j];
                        }
                    }
                }
            }
        }
        return impr;
    }

    /**
     * written because of bug: wrong number of objects after filtering in the case of dublicates.
     */
    private void resetClusterIDs() {
        System.out.println("---resetClusterIDs-----");
        for (int i = 0; i < bestClusters.length; i++) {
            int count = 0;
            for (int j = 0; j < data.length; j++) {
                if (data[j].clusterID == bestClusters[i].ID) {
                    count++;
                }
            }
            bestClusters[i].numObj = count;
        }
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            int count = 0;
            for (int j = 0; j < data.length; j++) {
                if (data[j].clusterID == bestNoiseClusters[i].ID) {
                    count++;
                }
            }
            bestNoiseClusters[i].numObj = count;
        }

    }

    /**
     * MODIFIED: method determines the cost of KMSplit. Writes global variables idsPCA and bestClustersPCA.
     * @return double: cost change: > 0 -> improvement of costs, < 0 -> cost decline. Double.NEGATIVEINFINITY if cluster is too small for splitting.
     * @param splitIndex index of cluster to be split
     */
    private double splitClusterPCA(int splitIndex) {
        double impr = 0.0;
        //System.out.println("split cluster " + bestClusters[splitIndex].ID);
        if (bestClusters[splitIndex].numObj < minNumObj) { //cluster is too small for splitting
            maxClusterID = maxClusterID + 2;
            return Double.NEGATIVE_INFINITY;
        }
        int[] currentID1 = new int[numObj]; //clusterIDs for cluster and noise cluster if filtered
        int[] currentID2;
        boolean improvement = false;
        boolean filtered1 = false; //k-means cluster 1 filtered?
        boolean filtered2 = false;
        //int splitIndex = chooseSplitCluster();
        //get all objects of the cluster to split
        double costOld = bestClusters[splitIndex].cost + bestNoiseClusters[splitIndex].cost;
        DataObject[] cobj = db.getObjects(bestClusters[splitIndex].ID); // hier obj kopieren und ClusterID auf 0 setzen!
        DataObject[] nobj = db.getObjects(bestNoiseClusters[splitIndex].ID);
        int numAll = cobj.length + nobj.length;
        DataObject[] all = new DataObject[numAll];
        for (int i = 0; i < numAll; i++) {
            if (i < cobj.length) {
                all[i] = cobj[i];
            } else {
                all[i] = nobj[i - cobj.length];
            }
        }
        //        START: K-MEANS
        DB kmDb = new DB(all);
        KMeans k = new KMeans(2, kmDb);
        DataObject[] newIDs = k.run(); //ClusterIDs 1 and 2
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = kmDb.getObjects(1);
        if (id_1.length < minNumObj) { //MODIFIED: NO CLUSTERS < minNumObj
            maxClusterID = maxClusterID + 2;
            return Double.NEGATIVE_INFINITY;
        }
//         END: K-MEANS
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for (int i = 0; i < id_1.length; i++) {
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap<Integer, Integer> h_1 = new HashMap<Integer, Integer>(id_1.length);
        for (int i = 0; i < id_1.length; i++) {
            h_1.put(internalIndex_1[i], numbers_1[i]);
        }
        for (int i = 0; i < id_1.length; i++) {
            id_1[i].number = internalIndex_1[i];
        }
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        IcaCluster[] clusters1 = getBestFilteredClusters(id_1, false, false);
        //change ID of assigned Objects to the true object IDs
        if (clusters1[0].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters1[0].assignedObjects.length; i++) {
                clusters1[0].assignedObjects[i] = h_1.get(clusters1[0].assignedObjects[i]);
            }
        }

        currentID1 = aktClusterID;
        clusters1[0].ID = bestClusters[splitIndex].ID;
        clusters1[1].ID = bestNoiseClusters[splitIndex].ID;
        if (clusters1[1].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters1[1].assignedObjects.length; i++) {
                clusters1[1].assignedObjects[i] = h_1.get(clusters1[1].assignedObjects[i]);
            }
        }


////objects assigned to clusterID 2 by kMeans
        //K-MEANS
        DataObject[] id_2 = kmDb.getObjects(2);
        if (id_2.length < minNumObj) {
            maxClusterID = maxClusterID + 2;
            return Double.NEGATIVE_INFINITY;
        }
        //K-MEANS
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for (int i = 0; i < id_2.length; i++) {
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap<Integer, Integer> h_2 = new HashMap<Integer, Integer>(id_2.length);
        for (int i = 0; i < id_2.length; i++) {
            h_2.put(internalIndex_2[i], numbers_2[i]);
        }
        for (int i = 0; i < id_2.length; i++) {
            id_2[i].number = internalIndex_2[i];
        }
        IcaCluster[] clusters2 = getBestFilteredClusters(id_2, false, false);
        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;
        clusters2[1].ID = maxClusterID + 3;
        if (clusters2[0].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters2[0].assignedObjects.length; i++) {
                clusters2[0].assignedObjects[i] = h_2.get(clusters2[0].assignedObjects[i]);
            }
        }
        if (clusters2[1].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters2[1].assignedObjects.length; i++) {
                clusters1[1].assignedObjects[i] = h_2.get(clusters1[1].assignedObjects[i]);
            }
        }

        //update current clusters
        for (int i = 0; i < bestClusters.length; i++) {
            if (i != splitIndex) {
                clusters.add(bestClusters[i]);
            } else {
                clusters.add(clusters1[0]);
            }
        }
        clusters.add(clusters2[0]);
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            if (i != splitIndex) {
                noiseClusters.add(bestNoiseClusters[i]);
            } else {
                noiseClusters.add(clusters1[1]);
            }
        }
        noiseClusters.add(clusters2[1]);
        //determine saved cost
        impr = savedCost(false);
        //write global variables bestClustersPCA and bestNoiseClustersPCA
        //bestClustersPCA, bestNoiseClustersPCA
        IcaCluster[] newClusters = new IcaCluster[clusters.size()];
        IcaCluster[] newNoiseClusters = new IcaCluster[clusters.size()];
        for (int i = 0; i < clusters.size(); i++) {
            newClusters[i] = (IcaCluster) clusters.elementAt(i);
            int aktID = newClusters[i].ID;
            for (int j = 0; j < noiseClusters.size(); j++) {
                if (((IcaCluster) noiseClusters.elementAt(j)).ID == aktID + 1) {
                    newNoiseClusters[i] = (IcaCluster) noiseClusters.elementAt(j);
                }
            }
            if (newNoiseClusters[i] == null) {
                newNoiseClusters[i] = new IcaCluster();
                newNoiseClusters[i].ID = aktID + 1;
            }
        }
        for (int i = 0; i < noiseClusters.size(); i++) {
            newNoiseClusters[i] = (IcaCluster) noiseClusters.elementAt(i);
        }
        bestClustersPCA = newClusters;
        bestNoiseClustersPCA = newNoiseClusters;
        //clusterIDS
        maxClusterID = maxClusterID + 2;
        int splittedID = bestClusters[splitIndex].ID;

        for (int i = 0; i < currentID1.length; i++) {
            int index = ((Integer) h_1.get(i)).intValue();
            if (currentID1[i] == 0) {
                idsPCA[index] = splittedID;
            } else {
                idsPCA[index] = splittedID + 1;
            }
        }
        for (int i = 0; i < currentID2.length; i++) {
            int index = ((Integer) h_2.get(i)).intValue();
            if (currentID2[i] == 0) {
                idsPCA[index] = maxClusterID;
            } else {
                idsPCA[index] = maxClusterID + 1;
            }
        }
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        return impr;
    }

    /**
    ADDED FOR OUTLIER DETECTION: split cluster also if this results in too small clusters
     */
    private double splitClusterPCAOutlier(int splitIndex) {
        double impr = 0.0;
        //System.out.println("split cluster " + bestClusters[splitIndex].ID);
        int[] currentID1 = new int[numObj]; //clusterIDs for cluster and noise cluster if filtered
        int[] currentID2;
        boolean improvement = false;
        boolean filtered1 = false; //k-means cluster 1 filtered?
        boolean filtered2 = false;
        //int splitIndex = chooseSplitCluster();
        //get all objects of the cluster to split
        double costOld = bestClusters[splitIndex].cost + bestNoiseClusters[splitIndex].cost;
        DataObject[] cobj = db.getObjects(bestClusters[splitIndex].ID); // hier obj kopieren und ClusterID auf 0 setzen!
        DataObject[] nobj = db.getObjects(bestNoiseClusters[splitIndex].ID);
        int numAll = cobj.length + nobj.length;
        DataObject[] all = new DataObject[numAll];
        for (int i = 0; i < numAll; i++) {
            if (i < cobj.length) {
                all[i] = cobj[i];
            } else {
                all[i] = nobj[i - cobj.length];
            }
        }
        //        START: K-MEANS
        DB kmDb = new DB(all);
        KMeans k = new KMeans(2, kmDb);
        DataObject[] newIDs = k.run(); //ClusterIDs 1 and 2
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = kmDb.getObjects(1);
//         END: K-MEANS
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for (int i = 0; i < id_1.length; i++) {
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap<Integer, Integer> h_1 = new HashMap<Integer, Integer>(id_1.length);
        for (int i = 0; i < id_1.length; i++) {
            h_1.put(internalIndex_1[i], numbers_1[i]);
        }
        for (int i = 0; i < id_1.length; i++) {
            id_1[i].number = internalIndex_1[i];
        }
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        IcaCluster[] clusters1 = getBestFilteredClusters(id_1, false, false);
        if (clusters1[0].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters1[0].assignedObjects.length; i++) {
                clusters1[0].assignedObjects[i] = h_1.get(clusters1[0].assignedObjects[i]);
            }
        }
        if (clusters1[1].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters1[1].assignedObjects.length; i++) {
                clusters1[1].assignedObjects[i] = h_1.get(clusters1[1].assignedObjects[i]);
            }
        }

        currentID1 = aktClusterID;
        clusters1[0].ID = bestClusters[splitIndex].ID;
        clusters1[1].ID = bestNoiseClusters[splitIndex].ID;

////objects assigned to clusterID 2 by kMeans
        //K-MEANS
        DataObject[] id_2 = kmDb.getObjects(2);
        //K-MEANS
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for (int i = 0; i < id_2.length; i++) {
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap<Integer, Integer> h_2 = new HashMap<Integer, Integer>(id_2.length);
        for (int i = 0; i < id_2.length; i++) {
            h_2.put(internalIndex_2[i], numbers_2[i]);
        }
        for (int i = 0; i < id_2.length; i++) {
            id_2[i].number = internalIndex_2[i];
        }
        IcaCluster[] clusters2 = getBestFilteredClusters(id_2, false, false);
        if (clusters2[0].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters2[0].assignedObjects.length; i++) {
                clusters2[0].assignedObjects[i] = h_2.get(clusters2[0].assignedObjects[i]);
            }
        }
        if (clusters2[1].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters2[1].assignedObjects.length; i++) {
                clusters2[1].assignedObjects[i] = h_2.get(clusters2[1].assignedObjects[i]);
            }
        }

        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;
        clusters2[1].ID = maxClusterID + 3;
        //update current clusters
        for (int i = 0; i < bestClusters.length; i++) {
            if (i != splitIndex) {
                clusters.add(bestClusters[i]);
            } else {
                clusters.add(clusters1[0]);
            }
        }
        clusters.add(clusters2[0]);
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            if (i != splitIndex) {
                noiseClusters.add(bestNoiseClusters[i]);
            } else {
                noiseClusters.add(clusters1[1]);
            }
        }
        noiseClusters.add(clusters2[1]);
        //determine saved cost
        impr = savedCost(false);
        //write global variables bestClustersPCA and bestNoiseClustersPCA
        //bestClustersPCA, bestNoiseClustersPCA
        IcaCluster[] newClusters = new IcaCluster[clusters.size()];
        IcaCluster[] newNoiseClusters = new IcaCluster[clusters.size()];
        for (int i = 0; i < clusters.size(); i++) {
            newClusters[i] = (IcaCluster) clusters.elementAt(i);
            int aktID = newClusters[i].ID;
            for (int j = 0; j < noiseClusters.size(); j++) {
                if (((IcaCluster) noiseClusters.elementAt(j)).ID == aktID + 1) {
                    newNoiseClusters[i] = (IcaCluster) noiseClusters.elementAt(j);
                }
            }
            if (newNoiseClusters[i] == null) {
                newNoiseClusters[i] = new IcaCluster();
                newNoiseClusters[i].ID = aktID + 1;
            }
        }
        for (int i = 0; i < noiseClusters.size(); i++) {
            newNoiseClusters[i] = (IcaCluster) noiseClusters.elementAt(i);
        }
        bestClustersPCA = newClusters;
        bestNoiseClustersPCA = newNoiseClusters;
        //clusterIDS
        maxClusterID = maxClusterID + 2;
        int splittedID = bestClusters[splitIndex].ID;
//TEST: check here IDsPCA
        //checkClusterIDs(idsPCA);
        //TEST
        for (int i = 0; i < currentID1.length; i++) {
            int index = ((Integer) h_1.get(i)).intValue();
            if (currentID1[i] == 0) {
                idsPCA[index] = splittedID;
            } else {
                idsPCA[index] = splittedID + 1;
            }
        }
        for (int i = 0; i < currentID2.length; i++) {
            int index = ((Integer) h_2.get(i)).intValue();
            if (currentID2[i] == 0) {
                idsPCA[index] = maxClusterID;
            } else {
                idsPCA[index] = maxClusterID + 1;
            }
        }
        //TEST
        //checkClusterIDs(idsPCA);
        //TEST
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        return impr;
    }

    /**
     * MODIFIED: method determines the cost of ICASplit. Writes global variables idsICA and bestClustersICA.
     * @return double: cost change: > 0 -> improvement of costs, < 0 -> cost decline. Double.NEGATIVEINFINITY if cluster is too small for splitting.
     * @param splitIndex index of cluster to be split
     */
    private double splitClusterICA(int splitIndex) {
        double impr = 0.0;
        //System.out.println("split cluster " + bestClusters[splitIndex].ID);
        if (bestClusters[splitIndex].numObj < minNumObj) { //cluster is too small for splitting
            maxClusterID = maxClusterID + 2;
            return Double.NEGATIVE_INFINITY;
        }
        int[] currentID1 = new int[numObj]; //clusterIDs for cluster and noise cluster if filtered
        int[] currentID2;
        boolean improvement = false;
        boolean filtered1 = false; //k-means cluster 1 filtered?
        boolean filtered2 = false;
        //int splitIndex = chooseSplitCluster();
        //get all objects of the cluster to split
        double costOld = bestClusters[splitIndex].cost + bestNoiseClusters[splitIndex].cost;
        DataObject[] cobj = db.getObjects(bestClusters[splitIndex].ID); // hier obj kopieren und ClusterID auf 0 setzen!
        DataObject[] nobj = db.getObjects(bestNoiseClusters[splitIndex].ID);
        int numAll = cobj.length + nobj.length;
        DataObject[] all = new DataObject[numAll];
        for (int i = 0; i < numAll; i++) {
            if (i < cobj.length) {
                all[i] = cobj[i];
            } else {
                all[i] = nobj[i - cobj.length];
            }
        }

        //MODIFIED: USE ICASPLIT INSTEAD OF ORDINARY K-MEANS
        //START: ICASPLIT
        IcaSplit is = new IcaSplit(2, all);
        DataObject[] newIDs = is.run(); //ClusterIDs 1 and 2
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = is.getObjects(1);
        if (id_1.length < minNumObj) {//MODIFIED: NO CLUSTERS < minNumObj
            maxClusterID = maxClusterID + 2;
            return Double.NEGATIVE_INFINITY;
        }
////        //END: ICASPLIT
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for (int i = 0; i < id_1.length; i++) {
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap<Integer, Integer> h_1 = new HashMap<Integer, Integer>(id_1.length);
        for (int i = 0; i < id_1.length; i++) {
            h_1.put(internalIndex_1[i], numbers_1[i]);
        }
        for (int i = 0; i < id_1.length; i++) {
            id_1[i].number = internalIndex_1[i];
        }
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        IcaCluster[] clusters1 = getBestFilteredClusters(id_1, false, false);
        if (clusters1[0].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters1[0].assignedObjects.length; i++) {
                clusters1[0].assignedObjects[i] = h_1.get(clusters1[0].assignedObjects[i]);
            }
        }
        if (clusters1[1].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters1[1].assignedObjects.length; i++) {
                clusters1[1].assignedObjects[i] = h_1.get(clusters1[1].assignedObjects[i]);
            }
        }
        currentID1 = aktClusterID;
        clusters1[0].ID = bestClusters[splitIndex].ID;
        clusters1[1].ID = bestNoiseClusters[splitIndex].ID;

////objects assigned to clusterID 2 by kMeans
//        //ICASPLIT
        DataObject[] id_2 = is.getObjects(2);
        if (id_2.length < minNumObj) {
            maxClusterID = maxClusterID + 2;
            return Double.NEGATIVE_INFINITY;
        }
////        //ICASPLIT
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for (int i = 0; i < id_2.length; i++) {
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap<Integer, Integer> h_2 = new HashMap<Integer, Integer>(id_2.length);
        for (int i = 0; i < id_2.length; i++) {
            h_2.put(internalIndex_2[i], numbers_2[i]);
        }
        for (int i = 0; i < id_2.length; i++) {
            id_2[i].number = internalIndex_2[i];
        }
        IcaCluster[] clusters2 = getBestFilteredClusters(id_2, false, false);
        if (clusters2[0].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters2[0].assignedObjects.length; i++) {
                clusters2[0].assignedObjects[i] = h_2.get(clusters2[0].assignedObjects[i]);
            }
        }
        if (clusters2[1].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters2[1].assignedObjects.length; i++) {
                clusters2[1].assignedObjects[i] = h_2.get(clusters2[1].assignedObjects[i]);
            }
        }
        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;
        clusters2[1].ID = maxClusterID + 3;
        //update current clusters
        for (int i = 0; i < bestClusters.length; i++) {
            if (i != splitIndex) {
                clusters.add(bestClusters[i]);
            } else {
                clusters.add(clusters1[0]);
            }
        }
        clusters.add(clusters2[0]);
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            if (i != splitIndex) {
                noiseClusters.add(bestNoiseClusters[i]);
            } else {
                noiseClusters.add(clusters1[1]);
            }
        }
        noiseClusters.add(clusters2[1]);
        //determine saved cost
        impr = savedCost(false);
        //write global variables bestClustersICA and bestNoiseClustersICA
        //bestClustersICA, bestNoiseClustersICA
        IcaCluster[] newClusters = new IcaCluster[clusters.size()];
        IcaCluster[] newNoiseClusters = new IcaCluster[clusters.size()];
        for (int i = 0; i < clusters.size(); i++) {
            newClusters[i] = (IcaCluster) clusters.elementAt(i);
            int aktID = newClusters[i].ID;
            for (int j = 0; j < noiseClusters.size(); j++) {
                if (((IcaCluster) noiseClusters.elementAt(j)).ID == aktID + 1) {
                    newNoiseClusters[i] = (IcaCluster) noiseClusters.elementAt(j);
                }
            }
            if (newNoiseClusters[i] == null) {
                newNoiseClusters[i] = new IcaCluster();
                newNoiseClusters[i].ID = aktID + 1;
            }
        }
        for (int i = 0; i < noiseClusters.size(); i++) {
            newNoiseClusters[i] = (IcaCluster) noiseClusters.elementAt(i);
        }
        bestClustersICA = newClusters;
        bestNoiseClustersICA = newNoiseClusters;
        //clusterIDS
        maxClusterID = maxClusterID + 2;
        int splittedID = bestClusters[splitIndex].ID;

        for (int i = 0; i < currentID1.length; i++) {
            int index = ((Integer) h_1.get(i)).intValue();
            if (currentID1[i] == 0) {
                idsICA[index] = splittedID;
            } else {
                idsICA[index] = splittedID + 1;
            }
        }
        for (int i = 0; i < currentID2.length; i++) {
            int index = ((Integer) h_2.get(i)).intValue();
            if (currentID2[i] == 0) {
                idsICA[index] = maxClusterID;
            } else {
                idsICA[index] = maxClusterID + 1;
            }
        }

        clusters.removeAllElements();
        noiseClusters.removeAllElements();

        return impr;
    }

    /**
     * ADDED FOR OUTLIER DETECTION: always splits the cluster, regardless if too small clusters occur.
     */
    private double splitClusterICAOutlier(int splitIndex) {
        double impr = 0.0;
        //System.out.println("split cluster " + bestClusters[splitIndex].ID);
        int[] currentID1 = new int[numObj]; //clusterIDs for cluster and noise cluster if filtered
        int[] currentID2;
        boolean improvement = false;
        boolean filtered1 = false; //k-means cluster 1 filtered?
        boolean filtered2 = false;
        //int splitIndex = chooseSplitCluster();
        //get all objects of the cluster to split
        double costOld = bestClusters[splitIndex].cost + bestNoiseClusters[splitIndex].cost;
        DataObject[] cobj = db.getObjects(bestClusters[splitIndex].ID); // hier obj kopieren und ClusterID auf 0 setzen!
        DataObject[] nobj = db.getObjects(bestNoiseClusters[splitIndex].ID);
        int numAll = cobj.length + nobj.length;
        DataObject[] all = new DataObject[numAll];
        for (int i = 0; i < numAll; i++) {
            if (i < cobj.length) {
                all[i] = cobj[i];
            } else {
                all[i] = nobj[i - cobj.length];
            }
        }

        //MODIFIED: USE ICASPLIT INSTEAD OF ORDINARY K-MEANS
        //START: ICASPLIT
        IcaSplit is = new IcaSplit(2, all);
        DataObject[] newIDs = is.run(); //ClusterIDs 1 and 2
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = is.getObjects(1);
////        //END: ICASPLIT
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for (int i = 0; i < id_1.length; i++) {
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap<Integer, Integer> h_1 = new HashMap<Integer, Integer>(id_1.length);
        for (int i = 0; i < id_1.length; i++) {
            h_1.put(internalIndex_1[i], numbers_1[i]);
        }
        for (int i = 0; i < id_1.length; i++) {
            id_1[i].number = internalIndex_1[i];
        }
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        IcaCluster[] clusters1 = getBestFilteredClusters(id_1, false, false);
        if (clusters1[0].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters1[0].assignedObjects.length; i++) {
                clusters1[0].assignedObjects[i] = h_1.get(clusters1[0].assignedObjects[i]);
            }
        }
        if (clusters1[1].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters1[1].assignedObjects.length; i++) {
                clusters1[1].assignedObjects[i] = h_1.get(clusters1[1].assignedObjects[i]);
            }
        }
        currentID1 = aktClusterID;
        clusters1[0].ID = bestClusters[splitIndex].ID;
        clusters1[1].ID = bestNoiseClusters[splitIndex].ID;

////objects assigned to clusterID 2 by kMeans
//        //ICASPLIT
        DataObject[] id_2 = is.getObjects(2);
////        //ICASPLIT
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for (int i = 0; i < id_2.length; i++) {
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap<Integer, Integer> h_2 = new HashMap<Integer, Integer>(id_2.length);
        for (int i = 0; i < id_2.length; i++) {
            h_2.put(internalIndex_2[i], numbers_2[i]);
        }
        for (int i = 0; i < id_2.length; i++) {
            id_2[i].number = internalIndex_2[i];
        }
        IcaCluster[] clusters2 = getBestFilteredClusters(id_2, false, false);
        if (clusters2[0].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters2[0].assignedObjects.length; i++) {
                clusters2[0].assignedObjects[i] = h_2.get(clusters2[0].assignedObjects[i]);
            }
        }
        if (clusters2[1].assignedObjects != null) { //cluster not empty
            for (int i = 0; i < clusters2[1].assignedObjects.length; i++) {
                clusters2[1].assignedObjects[i] = h_2.get(clusters2[1].assignedObjects[i]);
            }
        }
        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;
        clusters2[1].ID = maxClusterID + 3;
        //update current clusters
        for (int i = 0; i < bestClusters.length; i++) {
            if (i != splitIndex) {
                clusters.add(bestClusters[i]);
            } else {
                clusters.add(clusters1[0]);
            }
        }
        clusters.add(clusters2[0]);
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            if (i != splitIndex) {
                noiseClusters.add(bestNoiseClusters[i]);
            } else {
                noiseClusters.add(clusters1[1]);
            }
        }
        noiseClusters.add(clusters2[1]);
        //determine saved cost
        impr = savedCost(false);
        //write global variables bestClustersICA and bestNoiseClustersICA
        //bestClustersICA, bestNoiseClustersICA
        IcaCluster[] newClusters = new IcaCluster[clusters.size()];
        IcaCluster[] newNoiseClusters = new IcaCluster[clusters.size()];
        for (int i = 0; i < clusters.size(); i++) {
            newClusters[i] = (IcaCluster) clusters.elementAt(i);
            int aktID = newClusters[i].ID;
            for (int j = 0; j < noiseClusters.size(); j++) {
                if (((IcaCluster) noiseClusters.elementAt(j)).ID == aktID + 1) {
                    newNoiseClusters[i] = (IcaCluster) noiseClusters.elementAt(j);
                }
            }
            if (newNoiseClusters[i] == null) {
                newNoiseClusters[i] = new IcaCluster();
                newNoiseClusters[i].ID = aktID + 1;
            }
        }
        for (int i = 0; i < noiseClusters.size(); i++) {
            newNoiseClusters[i] = (IcaCluster) noiseClusters.elementAt(i);
        }
        bestClustersICA = newClusters;
        bestNoiseClustersICA = newNoiseClusters;
        //clusterIDS
        maxClusterID = maxClusterID + 2;
        int splittedID = bestClusters[splitIndex].ID;

        for (int i = 0; i < currentID1.length; i++) {
            int index = ((Integer) h_1.get(i)).intValue();
            if (currentID1[i] == 0) {
                idsICA[index] = splittedID;
            } else {
                idsICA[index] = splittedID + 1;
            }
        }
        for (int i = 0; i < currentID2.length; i++) {
            int index = ((Integer) h_2.get(i)).intValue();
            if (currentID2[i] == 0) {
                idsICA[index] = maxClusterID;
            } else {
                idsICA[index] = maxClusterID + 1;
            }
        }

        clusters.removeAllElements();
        noiseClusters.removeAllElements();

        return impr;
    }

    /**
     * Determines the next cluster to split, splits the cluster and returns if improvment of cost results from this split.
     * @param aktTry aktTry: number of trys after minimum cost has been reached
     * @return boolean: improvement of cost
     */
    private boolean splitCluster(boolean splitAnyway, int splitIndex) {
        //System.out.println("split cluster " + bestClusters[splitIndex].ID);
        if (bestClusters[splitIndex].numObj < minNumObj) {
            return false;
        }
        int[] currentID1 = new int[numObj]; //clusterIDs for cluster and noise cluster if filtered
        int[] currentID2;
        boolean improvement = false;
        boolean filtered1 = false; //k-means cluster 1 filtered?
        boolean filtered2 = false;
        //int splitIndex = chooseSplitCluster();
        //get all objects of the cluster to split
        double costOld = bestClusters[splitIndex].cost + bestNoiseClusters[splitIndex].cost;
        DataObject[] cobj = db.getObjects(bestClusters[splitIndex].ID); // hier obj kopieren und ClusterID auf 0 setzen!
        DataObject[] nobj = db.getObjects(bestNoiseClusters[splitIndex].ID);
        int numAll = cobj.length + nobj.length;
        DataObject[] all = new DataObject[numAll];
        for (int i = 0; i < numAll; i++) {
            if (i < cobj.length) {
                all[i] = cobj[i];
            } else {
                all[i] = nobj[i - cobj.length];
            }
        }

        //MODIFIED: USE ICASPLIT INSTEAD OF ORDINARY K-MEANS
        //START: ICASPLIT
        IcaSplit is = new IcaSplit(2, all);
        DataObject[] newIDs = is.run(); //ClusterIDs 1 and 2
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = is.getObjects(1);
        if (id_1.length < minNumObj) //MODIFIED: NO CLUSTERS < minNumObj
        {
            return false;
        }
////        //END: ICASPLIT

//        START: K-MEANS
//        DB kmDb = new DB(all);
//        KMeans k = new KMeans(2, kmDb);
//        DataObject[] newIDs = k.run(); //ClusterIDs 1 and 2
//        //objects assigned to cluster 1 by kMeans
//        DataObject[] id_1 = kmDb.getObjects(1);
//        if(id_1.length < minNumObj) //MODIFIED: NO CLUSTERS < minNumObj
        //           return false;
//         END: K-MEANS

        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for (int i = 0; i < id_1.length; i++) {
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap h_1 = new HashMap(id_1.length);
        for (int i = 0; i < id_1.length; i++) {
            h_1.put(internalIndex_1[i], numbers_1[i]);
        }
        for (int i = 0; i < id_1.length; i++) {
            id_1[i].number = internalIndex_1[i];
        }
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        IcaCluster[] clusters1 = getBestFilteredClusters(id_1, false, false);
        currentID1 = aktClusterID;
        clusters1[0].ID = bestClusters[splitIndex].ID;
        clusters1[1].ID = bestNoiseClusters[splitIndex].ID;

////objects assigned to clusterID 2 by kMeans
//        //ICASPLIT
        DataObject[] id_2 = is.getObjects(2);
        if (id_2.length < minNumObj) {
            return false;
        }
////        //ICASPLIT

        //K-MEANS
//        DataObject[] id_2 = kmDb.getObjects(2);
//        if(id_2.length < minNumObj)
//            return false;
        //K-MEANS

        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for (int i = 0; i < id_2.length; i++) {
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap h_2 = new HashMap(id_2.length);
        for (int i = 0; i < id_2.length; i++) {
            h_2.put(internalIndex_2[i], numbers_2[i]);
        }
        for (int i = 0; i < id_2.length; i++) {
            id_2[i].number = internalIndex_2[i];
        }
        IcaCluster[] clusters2 = getBestFilteredClusters(id_2, false, false);
        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;
        clusters2[1].ID = maxClusterID + 3;
        //update current clusters
        for (int i = 0; i < bestClusters.length; i++) {
            if (i != splitIndex) {
                clusters.add(bestClusters[i]);
            } else {
                clusters.add(clusters1[0]);
            }
        }
        clusters.add(clusters2[0]);
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            if (i != splitIndex) {
                noiseClusters.add(bestNoiseClusters[i]);
            } else {
                noiseClusters.add(clusters1[1]);
            }
        }
        noiseClusters.add(clusters2[1]);
        //boolean improvement = true;
        //boolean improvement = evaluateCost(splitIndex, cl, noise);
        if (splitAnyway) {
            improvement = true;
        } else {
            improvement = improvement(true);
        }
        if (improvement) {
            //set new clusters and new noiseClusters as best clusters.
            IcaCluster[] newClusters = new IcaCluster[clusters.size()];
            IcaCluster[] newNoiseClusters = new IcaCluster[clusters.size()];
            for (int i = 0; i < clusters.size(); i++) {
                newClusters[i] = (IcaCluster) clusters.elementAt(i);
                int aktID = newClusters[i].ID;
                for (int j = 0; j < noiseClusters.size(); j++) {
                    if (((IcaCluster) noiseClusters.elementAt(j)).ID == aktID + 1) {
                        newNoiseClusters[i] = (IcaCluster) noiseClusters.elementAt(j);
                    }
                }
                if (newNoiseClusters[i] == null) {
                    newNoiseClusters[i] = new IcaCluster();
                    newNoiseClusters[i].ID = aktID + 1;
                }
            }
            for (int i = 0; i < noiseClusters.size(); i++) {
                newNoiseClusters[i] = (IcaCluster) noiseClusters.elementAt(i);
            }
            bestClusters = newClusters;
            bestNoiseClusters = newNoiseClusters;
            //reset bestCost
            bestCost = 0.0;
            for (int i = 0; i < bestClusters.length; i++) {
                bestCost += bestClusters[i].cost;
            }
            for (int i = 0; i < bestNoiseClusters.length; i++) {
                bestCost += bestNoiseClusters[i].cost;
            }

            maxClusterID = maxClusterID + 2;
            int splittedID = bestClusters[splitIndex].ID;

            for (int i = 0; i < currentID1.length; i++) {
                int index = ((Integer) h_1.get(i)).intValue();
                if (currentID1[i] == 0) {
                    data[index].clusterID = splittedID;
                } else {
                    data[index].clusterID = splittedID + 1;
                }
            }
            for (int i = 0; i < currentID2.length; i++) {
                int index = ((Integer) h_2.get(i)).intValue();
                if (currentID2[i] == 0) {
                    data[index].clusterID = maxClusterID;
                } else {
                    data[index].clusterID = maxClusterID + 1;
                }
            }
        }
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        if (improvement) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Try once to split all clusters and all noise clusters individually. Called after top-down has reached first cost minimum and before merging.
     * @return boolean: improvement for at least one cluster or noiseCluster
     */
    private boolean finalSplit() {
        boolean impr = false;
        Vector<IcaCluster> nonEmpty = new Vector<IcaCluster>();
        for (int i = 0; i < bestClusters.length; i++) {
            if (bestClusters[i].numObj > 0) {
                nonEmpty.add(bestClusters[i]);
            }
        }
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            if (bestNoiseClusters[i].numObj > 0) {
                nonEmpty.add(bestNoiseClusters[i]);
            }
        }
        boolean[] improvement = new boolean[nonEmpty.size()];
        IcaCluster[] actClusters = new IcaCluster[nonEmpty.size()];
        for (int i = 0; i < actClusters.length; i++) {
            actClusters[i] = (IcaCluster) nonEmpty.elementAt(i);
        }
        //relable noiseClusters and generate dummy noiseClusters to use the method splitCluster()
        int clusterCount = 0;
        for (int i = 0; i < actClusters.length; i++) {
            int oldClusterID = actClusters[i].ID;
            if (oldClusterID != clusterCount) {
                actClusters[i].ID = clusterCount;
                adjustClusterIDs(oldClusterID, clusterCount);
            }
            clusterCount = clusterCount + 2;
        }
        maxClusterID = clusterCount;
        /*
        VisuSimple vk = new VisuSimple(data);
        vk.setSize(600,600);
        vk.setLocation(200,0);
        vk.setVisible(true);
         */

        IcaCluster[] actNoise = new IcaCluster[actClusters.length];
        for (int i = 0; i < actNoise.length; i++) {
            actNoise[i] = new IcaCluster(actClusters[i].ID + 1);
        }
        bestClusters = actClusters;
        bestNoiseClusters = actNoise;
        for (int i = 0; i < improvement.length; i++) {
            improvement[i] = splitCluster(true, i);
            if (improvement[i]) {
                impr = true;
            }
        }
        if (impr) {
            System.out.println("IMPROVEMENT DUE TO FINAL SPLIT. NEW: " + bestClusters.length + " clusters");
        }
        return impr;
    }

    private void setNewBestClusters() {
        IcaCluster[] newBest = new IcaCluster[clusters.size()];
        IcaCluster[] newBestNoise = new IcaCluster[clusters.size()];
        for (int i = 0; i < clusters.size(); i++) {
            newBest[i] = (IcaCluster) clusters.elementAt(i);
            newBestNoise[i] = new IcaCluster(-1);    //dummy noise f�r merge
        }
        bestNoiseClusters = newBestNoise;
        bestClusters = newBest;
    }

    private void adjustClusterIDs(int oldID, int newID) {
        for (int i = 0; i < data.length; i++) {
            if (data[i].clusterID == oldID) {
                data[i].clusterID = newID;
            }
        }

    }

    //for the current clustering return the coding costs of all objects
    private Double[] getObjectCodingCosts() {
        Double[] c = new Double[data.length];
        for (int i = 0; i < bestClusters.length; i++) {
            for (int j = 0; j < bestClusters[i].assignedObjects.length; j++) {
                c[bestClusters[i].assignedObjects[j]] = bestClusters[i].codingCostObjects[j];
            }
        }
        return c;
    }

    private Double[] getObjectParamCosts() {
        Double[] c = new Double[data.length];
        for (int i = 0; i < bestClusters.length; i++) {
            for (int j = 0; j < bestClusters[i].assignedObjects.length; j++) {
                c[bestClusters[i].assignedObjects[j]] = bestClusters[i].parameterCost / bestClusters[i].numObj;
            }
        }
        return c;
    }

    private Double[] getObjectIdCosts() {
        Double[] c = new Double[data.length];
        for (int i = 0; i < bestClusters.length; i++) {
            for (int j = 0; j < bestClusters[i].assignedObjects.length; j++) {
                c[bestClusters[i].assignedObjects[j]] = bestClusters[i].idCost / bestClusters[i].numObj;
            }
        }
        return c;
    }

    //assign objects the cluster with cost minimum. Writes only clusterID. 30.9.08. 
    //parameter overallCost: rows: objects, cols, merges in hierarchy.
    public void assignPoints(double[][] overallCost) {
        //for each level of the hierarchy determine mean object coding cost and select minimum among levels to normalize the outlier factor
        int mergeSteps = overallCost[0].length;
        double meanMinObjectCost = Double.MAX_VALUE; //to normalize outlier factor store mean minimum coding cost
        int minCostIndex = -1;
        for (int i = 0; i < mergeSteps; i++) {
            double sum = 0;
            for (int j = 0; j < numObj; j++) {
                sum += overallCost[j][i];
            }
            if (sum < meanMinObjectCost) {
                meanMinObjectCost = sum;
                minCostIndex = i;
            }
        }
        meanMinObjectCost/=numObj;
        System.out.println("best mean cost per object: " + meanMinObjectCost + " at mergeStep: " + minCostIndex);

        //set cluster ID of Objects to internal clusterID (cluster ID at the corresponding hierarchy level + mergeSteps)
        try {
            System.out.println("---ASSIGN POINTS----");

            int[] countMinCostsObjects = new int[mergeSteps];
            //set cluster ID of Objects to internal clusterID (cluster ID at the corresponding hierarchy level + mergeSteps)
            double[][] hierarchyLevelsCost = new double[numObj][2];
            for (int i = 0; i < numObj; i++) {
                //System.out.println("object: " + i);
                double minCost = Double.MAX_VALUE;
                int minIndex = -1;
                for (int j = 0; j < mergeSteps; j++) {
                    // System.out.println(overallCost[i][j] + " " + minCost);
                    if (overallCost[i][j] < minCost) {
                        minCost = overallCost[i][j];
                        minIndex = j;
                    }
                }
                data[i].hierarchyLevel = minIndex; //0: fine, mergeSteps: coarse
                data[i].codingCost = minCost/meanMinObjectCost;
                hierarchyLevelsCost[i][0] = minIndex;
                hierarchyLevelsCost[i][1] = minCost/meanMinObjectCost;
                countMinCostsObjects[minIndex]++;
            }
            int numClusterings = 0;
            for (int i = 0; i < mergeSteps; i++) {
                if (countMinCostsObjects[i] > 0) {
                    numClusterings++;
                }
            }
            System.out.println(numClusterings);
            //write hierarchy level and mincost for each object as .mat
            MLDouble mlHierarchy = new MLDouble("hierarchy", hierarchyLevelsCost);

            ArrayList l = new ArrayList();
            l.add(mlHierarchy);

            String filename = "hierarchy.mat";
            new MatFileWriter(filename, l);
        } catch (IOException ex) {
            Logger.getLogger(OCI.class.getName()).log(Level.SEVERE, null, ex);
        }




    }

    //returns overall costs of the objects.
    public double[][] greedyMergeForOutlier(boolean verbose) {
        //write coding costs, id costs and parametercosts and overallcosts to .mat file

        int mergeStep = -1;
        double[][] res = new double[1][1];
        IO ea = new IO();
        try {
            if (verbose) {
                System.out.println("---GREEDY MERGE FOR OUTLIER---------");
            }
            Vector<Double[]> codingCostsObjects = new Vector<Double[]>();
            Vector<Double[]> idCostsObjects = new Vector<Double[]>();
            Vector<Double[]> paramCostsObjects = new Vector<Double[]>();
            Vector<Double> numClusters = new Vector<Double>();
            //double[][] paramCostsObjects = new double[data.length][];
            codingCostsObjects.addElement(getObjectCodingCosts());
            idCostsObjects.addElement(getObjectIdCosts());
            paramCostsObjects.addElement(getObjectParamCosts());
            boolean finished = false;
            Vector<IcaCluster> aktClusters = new Vector<IcaCluster>(); //current clusters and non-empty noiseClusters
            clusters.removeAllElements();
            noiseClusters.removeAllElements();
            for (int i = 0; i < bestClusters.length; i++) {
                aktClusters.add(bestClusters[i]);
            }
            clusters = aktClusters;

            //save the order of the clusters in the cost table by saving their ID
            int[] order = new int[aktClusters.size()];
            for (int i = 0; i < aktClusters.size(); i++) {
                order[i] = aktClusters.elementAt(i).ID;
            }
            if (verbose) {
                System.out.println("order of clusters in cost table by ID: ");
                for (int i = 0; i < order.length; i++) {
                    System.out.print(order[i] + "\t");
                }
                System.out.println();
            }
            IcaCluster[] intermediate = new IcaCluster[aktClusters.size()];
            double[][] costTable = new double[aktClusters.size()][aktClusters.size()];

            for (int i = 0; i < costTable.length; i++) {
                for (int j = 0; j < costTable.length; j++) {
                    costTable[i][j] = -1;
                }
            }
            for (int i = 0; i < aktClusters.size(); i++) {
                for (int j = 0; j < aktClusters.size(); j++) {
                    if (i != j && costTable[i][j] == -1) {
                        //cost needs to be determined
                        for (int k = 0; k < intermediate.length; k++) {
                            intermediate[k] = (IcaCluster) aktClusters.elementAt(k);
                        }
                        IcaCluster a = (IcaCluster) aktClusters.elementAt(i);
                        IcaCluster b = (IcaCluster) aktClusters.elementAt(j);
                        IcaCluster m = merge(a, b);
//                    if(verbose)
//                        System.out.println("merge: " + a.ID + " and " + b.ID);
                        aktClusters.remove(a);
                        aktClusters.remove(b);
                        aktClusters.add(m);
                        costTable[i][j] = savedCost(false);
                        aktClusters.removeAllElements();
                        for (int k = 0; k < intermediate.length; k++) {
                            aktClusters.add(intermediate[k]);
                        }
                    } else {
                        costTable[i][j] = 0.0;
                    }
                }
            }

            if (verbose) {
                System.out.println("cost table");
                for (int i = 0; i < costTable.length; i++) {
                    for (int j = 0; j < costTable.length; j++) {
                        System.out.print(costTable[i][j] + "\t");
                    }
                    System.out.println();
                }
            }
            while (aktClusters.size() > 1) {
                //determine the index of the clusters to be merged next, i.e. the index with maximum saved cost.
                double maxSaved = -Double.MAX_VALUE;
                int i_index = -1;
                int j_index = -1;
                for (int i = 0; i < costTable.length; i++) {
                    for (int j = 0; j < costTable.length; j++) {
                        if (i != j && costTable[i][j] > maxSaved) {
                            maxSaved = costTable[i][j];
                            i_index = i;
                            j_index = j;
                        }
                    }
                }

                if (verbose) {
                    System.out.println("index of clusters to be merged in cost table: i_index: " + i_index + " j_index: " + j_index + " maxSaved: " + maxSaved);
                }
                IcaCluster a = (IcaCluster) (aktClusters.elementAt(i_index));
                IcaCluster b = (IcaCluster) (aktClusters.elementAt(j_index));
                IcaCluster m = merge(a, b);
                aktClusters.remove(a);
                aktClusters.remove(b);
                aktClusters.add(Math.min(i_index, j_index), m);
                int minID = Math.min(a.ID, b.ID);
                int maxID = Math.max(a.ID, b.ID);
                adjustClusterIDs(maxID, minID);
                setNewBestClusters();
                mergeStep++;
                //write clustering as serializable
                //create Clustering object: constructor needs: bestclusters, ids
                int[] currentIDs = new int[data.length];
                for (int i = 0; i < currentIDs.length; i++) {
                    currentIDs[i] = data[i].clusterID;
                }
                //coding costs, paramCosts, idCosts, overallCosts
                Double[] aktCoding = getObjectCodingCosts();
                Double[] aktId = getObjectIdCosts();
                Double[] aktParam = getObjectParamCosts();
                double[] currentCodingCosts = new double[data.length];
                double[] currentParamCosts = new double[data.length];
                double[] currentIdCosts = new double[data.length];
                double[] currentOverallCosts = new double[data.length];
                for (int i = 0; i < data.length; i++) {
                    currentCodingCosts[i] = aktCoding[i];
                    currentParamCosts[i] = aktParam[i];
                    currentIdCosts[i] = aktId[i];
                    currentOverallCosts[i] = aktCoding[i] + aktParam[i] + aktId[i];
                }
                Clustering c = new Clustering(bestClusters.clone(), currentIDs, currentCodingCosts, currentParamCosts, currentIdCosts, currentOverallCosts, mergeStep);
                ea.writeClusteringAsSerializable(c, "clusterings");
                codingCostsObjects.addElement(aktCoding);
                idCostsObjects.addElement(aktId);
                paramCostsObjects.addElement(aktParam);
                //determine cost of the new cluster to all other clusters
                double[] mCosts = new double[aktClusters.size() + 1];
                //init with -1
                for (int i = 0; i < mCosts.length; i++) {
                    mCosts[i] = -1;
                }
                mCosts[Math.min(i_index, j_index)] = 0.0; //index of new cluster
                mCosts[Math.max(i_index, j_index)] = -Double.MAX_VALUE; //index of removed cluster
                IcaCluster[] temp = new IcaCluster[aktClusters.size()]; //save current Clusters
                for (int i = 0; i < temp.length; i++) {
                    temp[i] = (IcaCluster) (aktClusters.elementAt(i));
                }
                int checked = 0;
                for (int i = 0; i < mCosts.length; i++) {
                    if (mCosts[i] == -1) {
                        //cost needs to be determined
                        IcaCluster akt = (IcaCluster) aktClusters.elementAt(checked);
                        IcaCluster merged = merge(m, akt);
                        aktClusters.remove(m);
                        aktClusters.remove(akt);
                        aktClusters.add(merged);
                        mCosts[i] = savedCost(false);
                        aktClusters.removeAllElements();
                        for (int j = 0; j < temp.length; j++) {
                            aktClusters.addElement(temp[j]);
                        }
                        checked++;
                    }
                    if (mCosts[i] == 0.0) {
                        checked++;
                    }
                }
                //update cost table
                //copy old cost table into array costsNew
                double[][] costsNew = new double[costTable.length][costTable.length];
                for (int i = 0; i < costsNew.length; i++) {
                    for (int j = 0; j < costsNew.length; j++) {
                        costsNew[i][j] = costTable[i][j];
                    }
                }
                //set row and column of removed cluster to -Double.MaxValue
                int indexRemoved = Math.max(i_index, j_index);
                for (int i = 0; i < costsNew.length; i++) {
                    costsNew[indexRemoved][i] = -Double.MAX_VALUE;
                }
                for (int i = 0; i < costsNew.length; i++) {
                    costsNew[i][indexRemoved] = -Double.MAX_VALUE;
                }
                //set row and colum of new cluster to mCosts
                int indexNew = Math.min(i_index, j_index);
                for (int i = 0; i < costsNew.length; i++) {
                    costsNew[indexNew][i] = mCosts[i];
                }
                for (int i = 0; i < costsNew.length; i++) {
                    costsNew[i][indexNew] = mCosts[i];
                }
                //copy all values != -Double.MaxValue to array costTableNew
                double[][] costTableNew = new double[aktClusters.size()][aktClusters.size()];
                int written_i = 0;
                int written_j = 0;
                for (int i = 0; i < costsNew.length; i++) {
                    for (int j = 0; j < costsNew.length; j++) {
                        if (costsNew[i][j] != -Double.MAX_VALUE) {
                            costTableNew[written_i][written_j] = costsNew[i][j];
                            written_j++;
                            if (written_j == costTableNew.length) {
                                //row full
                                written_i++;
                                written_j = 0;
                            }
                        }
                    }
                }
                costTable = costTableNew;
                //print cost table
                order = new int[aktClusters.size()];
                for (int i = 0; i < aktClusters.size(); i++) {
                    order[i] = aktClusters.elementAt(i).ID;
                }
                if (verbose) {
                    System.out.println("order of clusters in cost table by ID: ");
                    for (int i = 0; i < order.length; i++) {
                        System.out.print(order[i] + "\t");
                    }
                    System.out.println();
                }
                if (verbose) {
                    System.out.println("cost table");
                    for (int i = 0; i < costTable.length; i++) {
                        for (int j = 0; j < costTable.length; j++) {
                            System.out.print(costTable[i][j] + "\t");
                        }
                        System.out.println();
                    }
                }
            }
            //write coding costs, id costs and parametercosts and overallcosts to .mat file
            int steps = codingCostsObjects.size();
            System.out.println(steps);
            double[][] coding = new double[numObj][steps];
            double[][] params = new double[numObj][steps];
            double[][] ids = new double[numObj][steps];
            double[][] overall = new double[numObj][steps];
            for (int i = 0; i < steps; i++) {
                Double[] aktcoding = codingCostsObjects.elementAt(i);
                Double[] aktparams = paramCostsObjects.elementAt(i);
                Double[] aktids = idCostsObjects.elementAt(i);
                for (int j = 0; j < numObj; j++) {
                    coding[j][i] = aktcoding[j];
                    params[j][i] = aktparams[j];
                    ids[j][i] = aktids[j];
                    overall[j][i] = aktcoding[j] + aktparams[j] + aktids[j];
                }
            }
            MLDouble mlCoding = new MLDouble("coding", coding);
            MLDouble mlParams = new MLDouble("params", params);
            MLDouble mlIds = new MLDouble("ids", ids);
            MLDouble mlOverall = new MLDouble("overall", overall);
            ArrayList l = new ArrayList();
            l.add(mlCoding);
            l.add(mlParams);
            l.add(mlIds);
            l.add(mlOverall);
            String filename = "outlier.mat";
            new MatFileWriter(filename, l);
            res = overall;
        } catch (IOException ex) {
            Logger.getLogger(OCI.class.getName()).log(Level.SEVERE, null, ex);
        }
        return res;
    }

    public void greedyMerge(int numTry, boolean verbose) {
        if (verbose) {
            System.out.println("---GREEDY MERGE, numTry: " + numTry);
        }
        int aktTry = 0;
        Vector<IcaCluster> aktClusters = new Vector<IcaCluster>(); //current clusters and non-empty noiseClusters
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        for (int i = 0; i < bestClusters.length; i++) {
            aktClusters.add(bestClusters[i]);
        }
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            if (bestNoiseClusters[i].numObj > 0) {
                bestNoiseClusters[i].cost += dim * 32 * 2;
                aktClusters.add(bestNoiseClusters[i]);
            }
        }
        clusters = aktClusters;
        //save Clusters and IDs of the last cost minimum
        int[] lastLocalMinIDs = new int[numObj];
        for (int i = 0; i < numObj; i++) {
            lastLocalMinIDs[i] = data[i].clusterID;
        }
        IcaCluster[] lastLocalMinClusters = new IcaCluster[clusters.size()];
        for (int i = 0; i < lastLocalMinClusters.length; i++) {
            lastLocalMinClusters[i] = (IcaCluster) clusters.elementAt(i);
        }
        bestCost = currentCost();
        boolean improvement = true;
        //determine cost table
        //save the order of the clusters in the cost table by saving their ID
        int[] order = new int[aktClusters.size()];
        for (int i = 0; i < aktClusters.size(); i++) {
            order[i] = aktClusters.elementAt(i).ID;
        }
        if (verbose) {
            System.out.println("order of clusters in cost table by ID: ");
            for (int i = 0; i < order.length; i++) {
                System.out.print(order[i] + "\t");
            }
            System.out.println();
        }
        IcaCluster[] intermediate = new IcaCluster[aktClusters.size()];
        double[][] costTable = new double[aktClusters.size()][aktClusters.size()];
        //init cost table with -1 to check if cost has already been computed
        for (int i = 0; i < costTable.length; i++) {
            for (int j = 0; j < costTable.length; j++) {
                costTable[i][j] = -1;
            }
        }
        for (int i = 0; i < aktClusters.size(); i++) {
            for (int j = 0; j < aktClusters.size(); j++) {
                if (i != j && costTable[i][j] == -1) { //cost needs to be determined
                    for (int k = 0; k < intermediate.length; k++) {
                        intermediate[k] = (IcaCluster) aktClusters.elementAt(k);
                    }
                    IcaCluster a = (IcaCluster) aktClusters.elementAt(i);
                    IcaCluster b = (IcaCluster) aktClusters.elementAt(j);
                    IcaCluster m = merge(a, b);
//                    if(verbose)
//                        System.out.println("merge: " + a.ID + " and " + b.ID);
                    aktClusters.remove(a);
                    aktClusters.remove(b);
                    aktClusters.add(m);
                    costTable[i][j] = savedCost(false);
                    aktClusters.removeAllElements();
                    for (int k = 0; k < intermediate.length; k++) {
                        aktClusters.add(intermediate[k]);
                    }
                } else {
                    costTable[i][j] = 0.0;
                }
            }
        }
        //print cost table
        if (verbose) {
            System.out.println("cost table");
            for (int i = 0; i < costTable.length; i++) {
                for (int j = 0; j < costTable.length; j++) {
                    System.out.print(costTable[i][j] + "\t");
                }
                System.out.println();
            }
        }
        while (improvement && aktClusters.size() > 1) {
            //determine the index of the clusters to be merged next, i.e. the index with maximum saved cost.
            double maxSaved = -Double.MAX_VALUE;
            int i_index = -1;
            int j_index = -1;
            for (int i = 0; i < costTable.length; i++) {
                for (int j = 0; j < costTable.length; j++) {
                    if (i != j && costTable[i][j] > maxSaved) {
                        maxSaved = costTable[i][j];
                        i_index = i;
                        j_index = j;
                    }
                }
            }

            if (verbose) {
                System.out.println("index of clusters to be merged in cost table: i_index: " + i_index + " j_index: " + j_index + " maxSaved: " + maxSaved);
            }
            if (maxSaved < 0 && aktTry >= numTry) //stop
            {
                improvement = false;
            } else { //improvement due to merging
                aktTry++;
                IcaCluster a = (IcaCluster) (aktClusters.elementAt(i_index));
                IcaCluster b = (IcaCluster) (aktClusters.elementAt(j_index));
                IcaCluster m = merge(a, b);
                aktClusters.remove(a);
                aktClusters.remove(b);
                aktClusters.add(Math.min(i_index, j_index), m);
                int minID = Math.min(a.ID, b.ID);
                int maxID = Math.max(a.ID, b.ID);
                adjustClusterIDs(maxID, minID);
                setNewBestClusters();
                //TEST
                //testMethodForCodingCostObjects();
                //testMethodForAssignedObjects();
                //TEST
                //determine cost of the new cluster to all other clusters
                double[] mCosts = new double[aktClusters.size() + 1];
                //init with -1
                for (int i = 0; i < mCosts.length; i++) {
                    mCosts[i] = -1;
                }
                mCosts[Math.min(i_index, j_index)] = 0.0; //index of new cluster
                mCosts[Math.max(i_index, j_index)] = -Double.MAX_VALUE; //index of removed cluster
                IcaCluster[] temp = new IcaCluster[aktClusters.size()]; //save current Clusters
                for (int i = 0; i < temp.length; i++) {
                    temp[i] = (IcaCluster) (aktClusters.elementAt(i));
                }
                int checked = 0;
                for (int i = 0; i < mCosts.length; i++) {
                    if (mCosts[i] == -1) { //cost needs to be determined
                        IcaCluster akt = (IcaCluster) aktClusters.elementAt(checked);
                        IcaCluster merged = merge(m, akt);
                        aktClusters.remove(m);
                        aktClusters.remove(akt);
                        aktClusters.add(merged);
                        mCosts[i] = savedCost(false);
                        aktClusters.removeAllElements();
                        for (int j = 0; j < temp.length; j++) {
                            aktClusters.addElement(temp[j]);
                        }
                        checked++;
                    }
                    if (mCosts[i] == 0.0) {
                        checked++;
                    }
                }
                //update cost table
                //copy old cost table into arry costsNew
                double[][] costsNew = new double[costTable.length][costTable.length];
                for (int i = 0; i < costsNew.length; i++) {
                    for (int j = 0; j < costsNew.length; j++) {
                        costsNew[i][j] = costTable[i][j];
                    }
                }
                //set row and column of removed cluster to -Double.MaxValue
                int indexRemoved = Math.max(i_index, j_index);
                for (int i = 0; i < costsNew.length; i++) {
                    costsNew[indexRemoved][i] = -Double.MAX_VALUE;
                }
                for (int i = 0; i < costsNew.length; i++) {
                    costsNew[i][indexRemoved] = -Double.MAX_VALUE;
                }
                //set row and colum of new cluster to mCosts
                int indexNew = Math.min(i_index, j_index);
                for (int i = 0; i < costsNew.length; i++) {
                    costsNew[indexNew][i] = mCosts[i];
                }
                for (int i = 0; i < costsNew.length; i++) {
                    costsNew[i][indexNew] = mCosts[i];
                }
                //copy all values != -Double.MaxValue to array costTableNew
                double[][] costTableNew = new double[aktClusters.size()][aktClusters.size()];
                int written_i = 0;
                int written_j = 0;
                for (int i = 0; i < costsNew.length; i++) {
                    for (int j = 0; j < costsNew.length; j++) {
                        if (costsNew[i][j] != -Double.MAX_VALUE) {
                            costTableNew[written_i][written_j] = costsNew[i][j];
                            written_j++;
                            if (written_j == costTableNew.length) { //row full
                                written_i++;
                                written_j = 0;
                            }
                        }
                    }
                }
                costTable = costTableNew;
                //print cost table
                order = new int[aktClusters.size()];
                for (int i = 0; i < aktClusters.size(); i++) {
                    order[i] = aktClusters.elementAt(i).ID;
                }
                if (verbose) {
                    System.out.println("order of clusters in cost table by ID: ");
                    for (int i = 0; i < order.length; i++) {
                        System.out.print(order[i] + "\t");
                    }
                    System.out.println();
                }
                if (verbose) {
                    System.out.println("cost table");
                    for (int i = 0; i < costTable.length; i++) {
                        for (int j = 0; j < costTable.length; j++) {
                            System.out.print(costTable[i][j] + "\t");
                        }
                        System.out.println();
                    }
                }
                if (maxSaved > 0) { //local cost minimum: save clusters and IDs as lastlocalMinClusters and lastlocalMinIDs
                    for (int i = 0; i < numObj; i++) {
                        lastLocalMinIDs[i] = data[i].clusterID;
                    }
                    lastLocalMinClusters = bestClusters;
                }
            } //else (improvement due to merging
        } //while loop
    }

    private void checkClusterIDs(int[] ids) {
        System.out.println("---checkClusterIDs-----");
        Vector<Integer> id = new Vector<Integer>();
        for (int i = 0; i < ids.length; i++) {
            if (!id.contains(ids[i])) {
                id.addElement(ids[i]);
            }
        }
        System.out.println("CURRENT CLUSTER IDS:");
        for (int i = 0; i < id.size(); i++) {
            System.out.print(id.elementAt((i)) + " ");
            int count = 0;
            for (int j = 0; j < ids.length; j++) {
                if (ids[j] == id.elementAt(i)) {
                    count++;
                }
            }
            System.out.print(count);
            System.out.println();
        }
    }

    private void checkClusterIDs() {
//collect current clusterIDs
        System.out.println("---checkClusterIDs-----");
        Vector<Integer> id = new Vector<Integer>();
        for (int i = 0; i < bestClusters.length; i++) {
            id.addElement(bestClusters[i].ID);
        }
        System.out.println("CURRENT CLUSTER IDS:");
        for (int i = 0; i < id.size(); i++) {
            System.out.print(id.elementAt((i)) + " ");
            int count = 0;
            for (int j = 0; j < data.length; j++) {
                if (data[j].clusterID == id.elementAt(i)) {
                    count++;
                }
            }
            System.out.print(count);
            System.out.println();
        }
    }

    private IcaCluster merge(IcaCluster c, IcaCluster d) {
        boolean verbose = false;
        if (verbose) {
            System.out.println("merge clusters: " + c.ID + " " + d.ID);
        }
        //determine best model: rotated or not and TODO: best distribution
        DataObject[] mObj = getMergedObjects(c, d);
        if (verbose) {
            System.out.println("c.numObj: " + c.numObj + " d.numObj: " + d.numObj + " m.numObj: " + mObj.length);
        }
        int minID = Math.min(c.ID, d.ID);
        int maxID = Math.max(c.ID, d.ID);
        if (minID == maxID) {
            System.err.println("trying to merge clusters with the same id: " + minID);
        }
        if (mObj.length < minNumObj) {
            if (verbose) {
                System.out.println("Cluster too small for ica");
            }
            IcaCost ic = new IcaCost(dim);
            IcaCluster m = ic.mergedClusterNoRot(mObj);
            m.numObj = mObj.length;
            m.ID = minID;
            return m;
        }
        IcaCost icaC = new IcaCost(dim);
        IcaCluster m = icaC.selectBestCluster(mObj);
        //m.cost += paramCost;
        m.ID = minID;
        return m;
    }

    private DataObject[] getObjects(IcaCluster c) {
        DataObject[] members = new DataObject[c.numObj];
        int counter = 0;
        for (int j = 0; j < data.length; j++) {
            if (data[j].clusterID == c.ID) {
                members[counter] = data[j];
                counter++;
            }
        }
        return members;
    }

    private DataObject[] getMergedObjects(IcaCluster c, IcaCluster d) {
        //DataObject[] members = new DataObject[c.numObj + d.numObj];
        DataObject[] members = new DataObject[countObj(c.ID, d.ID)];
        int counter = 0;
        for (int j = 0; j < data.length; j++) {
            if (data[j].clusterID == c.ID || data[j].clusterID == d.ID) {
                members[counter] = data[j];
                counter++;
            }
        }
        return members;
    }

    public int countObj(int cId, int dId) {
        int counter = 0;
        for (int i = 0; i < data.length; i++) {
            if (data[i].clusterID == cId || data[i].clusterID == dId) {
                counter++;
            }
        }
        return counter;
    }

    /**
     * Returns the current cost of bestClusters and bestNoiseClusters including ID cost
     * @return double: cost
     */
    private double currentCost() {
        boolean verbose = false;
        if (verbose) {
            System.out.println("currentCost");
        }
        double cost = 0.0;
        for (int i = 0; i < bestClusters.length; i++) {
            IcaCluster akt = bestClusters[i];
            if (akt.numObj > 0) {
                double costID = akt.numObj * Math.log((double) numObj / (double) akt.numObj) / Math.log(2);
                akt.idCost = costID;
                if (verbose) {
                    System.out.println("Cluster " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " CostID " + costID);
                }
                cost += costID;
                cost += bestClusters[i].cost;
            }
        }
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            IcaCluster akt = bestNoiseClusters[i];
            if (akt.numObj > 0) {
                double costID = akt.numObj * Math.log((double) numObj / (double) akt.numObj) / Math.log(2);
                akt.idCost = costID;
                if (verbose) {
                    System.out.println("Noise " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " CostID " + costID);
                }
                cost += costID;
                cost += bestNoiseClusters[i].cost;
            }
        }
        return cost;
    }

    /**
     * Checks if the costs are lower after one split. Cost for clusterID are counted here.
     * @return boolean: improvement of cost
     */
    private boolean improvement(boolean verbose) {
        double costNew = 0.0;
        double costOld = 0.0;
        double costIDNew = 0.0;
        double costIDOld = 0.0;
        int clusterCountNew = 0;
        int clusterCountOld = 0;
        if (verbose) {
            System.out.println("NEW:");
        }
        for (int i = 0; i < clusters.size(); i++) {
            IcaCluster akt = (IcaCluster) clusters.elementAt(i);
            if (clusters.elementAt(i).numObj > 0) {
                clusterCountNew++;
                double costID = akt.numObj * Math.log((double) numObj / (double) akt.numObj) / Math.log(2);
                if (verbose) {
                    System.out.println("cluster " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " costID: " + costID);
                }
                costNew += costID;
                costNew += akt.cost;
            } else if (verbose) {
                System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
            }
        }
        for (int i = 0; i < noiseClusters.size(); i++) {
            IcaCluster akt = (IcaCluster) noiseClusters.elementAt(i);
            if (noiseClusters.elementAt(i).numObj > 0) {
                double costID = akt.numObj * Math.log((double) numObj / (double) akt.numObj) / Math.log(2);
                if (verbose) {
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                }
                costNew += costID;
                clusterCountNew++;
                costNew += akt.cost;
            } else if (verbose) {
                System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
            }
        }
        if (verbose) {
            System.out.println("OLD:");
        }
        for (int i = 0; i < bestClusters.length; i++) {
            IcaCluster akt = bestClusters[i];
            if (akt.numObj > 0) {
                clusterCountOld++;
                double costID = akt.numObj * Math.log((double) numObj / (double) akt.numObj) / Math.log(2);
                costOld += costID;
                if (verbose) {
                    System.out.println("cluster " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " costID: " + costID);
                }
                costOld += bestClusters[i].cost;
            } else if (verbose) {
                System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
            }
        }
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            IcaCluster akt = bestNoiseClusters[i];
            if (akt.numObj > 0) {
                double costID = akt.numObj * Math.log((double) numObj / (double) akt.numObj) / Math.log(2);
                costOld += costID;
                clusterCountOld++;
                if (verbose) {
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                }
                costOld += bestNoiseClusters[i].cost;
            } else if (verbose) {
                System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
            }
        }
        if (costNew < costOld) {
            if (verbose) {
                System.out.println("improvement : cost old: " + costOld + " " + clusterCountOld + " clusters, cost new: " + " " + costNew + " " + clusterCountNew + " clusters.");
            }
            return true;
        } else {
            if (verbose) {
                System.out.println("no improvement : cost old: " + costOld + " " + clusterCountOld + " clusters, cost new: " + " " + costNew + " " + clusterCountNew + " clusters.");
            }
            return false;
        }
    }

    private double savedCost(boolean verbose) {
        double costNew = 0.0;
        double costOld = 0.0;
        double costIDNew = 0.0;
        double costIDOld = 0.0;
        int clusterCountNew = 0;
        int clusterCountOld = 0;
        double savedCost;
        if (verbose) {
            System.out.println("NEW:");
        }
        for (int i = 0; i < clusters.size(); i++) {
            IcaCluster akt = (IcaCluster) clusters.elementAt(i);
            if (clusters.elementAt(i).numObj > 0) {
                clusterCountNew++;
                double costID = akt.numObj * Math.log((double) numObj / (double) akt.numObj) / Math.log(2);
                akt.idCost = costID;
                if (verbose) {
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                }
                costNew += costID;
                costNew += akt.cost;
            } else if (verbose) {
                System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
            }
        }
        for (int i = 0; i < noiseClusters.size(); i++) {
            IcaCluster akt = (IcaCluster) noiseClusters.elementAt(i);
            if (noiseClusters.elementAt(i).numObj > 0) {
                double costID = akt.numObj * Math.log((double) numObj / (double) akt.numObj) / Math.log(2);
                akt.idCost = costID;
                if (verbose) {
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                }
                costNew += costID;
                clusterCountNew++;
                costNew += akt.cost;
            } else if (verbose) {
                System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
            }
        }
        if (verbose) {
            System.out.println("OLD:");
        }
        for (int i = 0; i < bestClusters.length; i++) {
            IcaCluster akt = bestClusters[i];
            if (akt.numObj > 0) {
                clusterCountOld++;
                double costID = akt.numObj * Math.log((double) numObj / (double) akt.numObj) / Math.log(2);
                akt.idCost = costID; //should be not necessary to set it again here
                costOld += costID;
                if (verbose) {
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                }
                costOld += bestClusters[i].cost;
            } else if (verbose) {
                System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
            }
        }
        for (int i = 0; i < bestNoiseClusters.length; i++) {
            IcaCluster akt = bestNoiseClusters[i];
            if (akt.numObj > 0) {
                double costID = akt.numObj * Math.log((double) numObj / (double) akt.numObj) / Math.log(2);
                akt.idCost = costID; //should be not necessary to set it again here
                costOld += costID;
                clusterCountOld++;
                if (verbose) {
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                }
                costOld += bestNoiseClusters[i].cost;
            } else if (verbose) {
                System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
            }
        }
        return costOld - costNew;
    }

    public void getCostInfo() {
        System.out.println("----getCostInfo()----");
        double overallID = 0.0;
        double overallCoding = 0.0;
        double overallParams = 0.0;
        for (int i = 0; i < bestClusters.length; i++) {
            System.out.println("cluster " + bestClusters[i].ID + " numObj: " + bestClusters[i].numObj + " coding costs: " + bestClusters[i].codingCost + " parameter costs: " + bestClusters[i].parameterCost + " id costs: " + bestClusters[i].idCost);
            //sanity check
            if (bestClusters[i].codingCost + bestClusters[i].parameterCost != bestClusters[i].cost) {
                System.err.println("wrong costs");
            }
            overallID += bestClusters[i].idCost;
            overallCoding += bestClusters[i].codingCost;
            overallParams += bestClusters[i].parameterCost;
        }
        System.out.println("overall costs of clustering: coding: " + overallCoding + " params: " + overallParams + " ids: " + overallID);
    }

    public void getCostInfoShort() {
        double overallID = 0.0;
        double overallCoding = 0.0;
        double overallParams = 0.0;
        for (int i = 0; i < bestClusters.length; i++) {
            //System.out.println("cluster " + bestClusters[i].ID + " numObj: " + bestClusters[i].numObj + " coding costs: " + bestClusters[i].codingCost + " parameter costs: " + bestClusters[i].parameterCost + " id costs: " + bestClusters[i].idCost);
            //sanity check
            if (bestClusters[i].codingCost + bestClusters[i].parameterCost != bestClusters[i].cost) {
                System.err.println("wrong costs");
            }
            overallID += bestClusters[i].idCost;
            overallCoding += bestClusters[i].codingCost;
            overallParams += bestClusters[i].parameterCost;
        }
        double overall = overallCoding + overallID + overallParams;
        System.out.println(bestClusters.length + " " + overallCoding + " " + overallParams + " " + overallID + " " + overall);
    }

    public int countClusterMembers(int clusterID) {
        int counter = 0;
        for (int i = 0; i < data.length; i++) {
            if (data[i].clusterID == clusterID) {
                counter++;
            }
        }
        return counter;
    }

    private int countClassMembers(int classID) {
        int clM = 0;
        for (int i = 0; i < data.length; i++) {
            if (data[i].classID == classID) {
                clM++;
            }
        }
        return clM;
    }

    private int countClassClusterMembers(int clusterID, int classID) {
        int clM = 0;
        for (int i = 0; i < data.length; i++) {
            if (data[i].classID == classID && data[i].clusterID == clusterID) {
                clM++;
            }
        }
        return clM;
    }

    private int countClasses() {
        Vector clCount = new Vector();
        for (int i = 0; i < data.length; i++) {
            Integer clID = (Integer) (data[i].classID);
            if (!clCount.contains(clID)) {
                clCount.add(clID);
            }
        }
        return clCount.size();
    }

    private void printPercentage() {
        int numClasses = countClasses();
        int[] classMembers = new int[numClasses];
        for (int i = 0; i < numClasses; i++) {
            classMembers[i] = countClassMembers(i + 1);
        }
        for (int i = 0; i < bestClusters.length; i++) {
            IcaCluster akt = bestClusters[i];
            System.out.println("cluster " + akt.ID);
            for (int j = 0; j < numClasses; j++) {
                //System.out.println("class " + (j+1) + " " + ((double)countClassClusterMembers(akt.ID, j+1)/(double)classMembers[j]) + " percent ");
                System.out.println("class " + (j + 1) + " " + countClassClusterMembers(akt.ID, j + 1));
            //System.out.println();
            }
            System.out.println();
        }
    }
}




